# pylint: disable=E1101
import numpy as np
from apdt.ml.tf_model import mlp_weight, MLP

try:
    import tensorflow as tf
except:
    pass

def deepsets_weight(**kwarg):
    '''Construct weight for deepsets.

    Parameter
    ---------
        name: str
            the variable name scope to be used.
        input_dim: int, default 10
            the number of input neurals
        output_dim: int, default 10
            the number of output neurals
        n_hidden: list of int, default 128
            the dimension of hidden states.
        mlp_n_hidden: list of int, default [128,128]
            the number of mlp hidden units, the length of list decide the number of layers.
    
    Return
    ------
        dict
            the dictionary of weight tensor.
    '''
    if 'name' not in kwarg:
        kwarg['name'] = 'mlp'
    if 'input_dim' not in kwarg:
        kwarg['input_dim'] = 10
    if 'output_dim' not in kwarg:
        kwarg['output_dim'] = 10
    if 'n_hidden' not in kwarg:
        kwarg['n_hidden'] = 128
    if 'mlp_n_hidden' not in kwarg:
        kwarg['mlp_n_hidden'] = [128, 128]   
    if 'trainable' not in kwarg:
        kwarg['trainable'] = True

    w = {}
    with tf.variable_scope(kwarg['name']):
        mlp_weight_encoder = mlp_weight(name="encoder_mlp", input_dim=kwarg['input_dim'], output_dim=kwarg['n_hidden'],
                                        n_hidden=kwarg['mlp_n_hidden'],trainable=kwarg['trainable'])
        mlp_weight_decoder = mlp_weight(name="decoder_mlp", input_dim=kwarg['n_hidden'], output_dim=kwarg['output_dim'],
                                        n_hidden=kwarg['mlp_n_hidden'],trainable=kwarg['trainable'])
    w.update({'pf0'+key: value for key, value in mlp_weight_encoder.items()})
    w.update({'pf1'+key: value for key, value in mlp_weight_decoder.items()})

    return w

def deepsets(x, axis, weights=None, name='deepsets', **kwarg):
    '''Map a set into one feature using deepsets.

    Parameter
    ---------
        x: tensor
            the input tensor
        axis: int
            along which axis is the set going
        weights: dict
            a dictionary of weights generated by apdt.ml.deepsets_weight
    
    Return
    ------
        tensor
            The output tensor
    
    '''
    
    if weights is None:
        if "output_dim" not in kwarg:
            raise Exception("output_dim should be given if weights is None.")
        weights = deepsets_weight(name=name, input_dim=x.shape[-1].value, output_dim=kwarg["output_dim"])
    
    mlp_weight_encoder = {key[3:]: value for key, value in weights.items() if key.startswith('pf0')}
    mlp_weight_decoder = {key[3:]: value for key, value in weights.items() if key.startswith('pf1')}
    with tf.variable_scope(name):
        a = MLP(x, weights=mlp_weight_encoder, name="encoder_mlp")
        h = tf.reduce_mean(a, axis)
        z = MLP(h, weights=mlp_weight_decoder, name="decoder_mlp")

    return z



'''
APDT: Machine Learning
--------------------------

Provide useful modeling and training interface for deep learning.

Check https://github.com/Zhiyuan-Wu/apdt for more information.
'''
from apdt.ml.general import DataSet, TFModel
from apdt.ml.tf_model import wavenet_weight, WaveNet, mlp_weight, MLP, lstm_weight, LSTM
from apdt.ml.tf_model import cnn1d_weight, cnn1d, cnn2d_weight, cnn2d
from apdt.ml.tf_model import Transformer_weight, Transformer
from apdt.ml.tf_model import _WaveNet
from apdt.ml.tf_model2 import deepsets_weight, deepsets
from apdt.ml.utils import stacked_window, batch_norm, layer_norm, pearson_corr, percentile, median, KL_diagonal_gaussian
from apdt.ml.special_dataset import sin_regression, ContinualDataset


import numpy as np
import itertools

try:
    import tensorflow as tf
except:
    pass

def batch_norm(x, training_flag, gamma=0.99, learnable=True, name='bn'):
    '''Apply batch norm to input x, i.e., turn each feature channel into a (learnable) gaussian distribution. 
    This function will introduce new in-trainable variables into models. 
    Its Useful to insert this function between linear mapping and activation to get magical improvements/degradation.

    Parameters
    ------
        x, tensor
            input time series of shape [batch_size, spatial_dim_1, ..., spatial_dim_S, feature_dim]
        training_flag, tensor
            a bool scalar tensor indicating if its training phase, usually TFModel.training
        gamma, float, default 0.99
            moving average decay weight.
        learnable, bool, default True
            whether to introduce learnable mean and var to target distribution.
        name, str, default "bn"
            the name prefix for new variables
    Returns
    ------
        tensor
            same shape with x
    Notes
    ------
        This function include in-trainable variables and maybe incompitable with current pretrain model loading mechanism.
    '''
    num_dims = x.shape.ndims
    D = x.shape[-1].value

    running_mean = tf.get_variable(name+'_running_mean', [D,], initializer=tf.constant_initializer(0.0),trainable=False)
    running_var = tf.get_variable(name+'_running_var', [D,], initializer=tf.constant_initializer(1.0),trainable=False)
    if learnable:
        learnable_mean = tf.get_variable(name+'_learnable_mean', [D,], initializer=tf.constant_initializer(0.0))
        learnable_var = tf.get_variable(name+'_learnable_var', [D,], initializer=tf.constant_initializer(1.0))
    mean, var = tf.nn.moments(x, [i for i in range(num_dims-1)])
    tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, tf.assign(running_mean, running_mean * gamma + mean * (1-gamma)))
    tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, tf.assign(running_var, running_var * gamma + var * (1-gamma)))
    mean, var = tf.cond(training_flag, lambda: (mean, var), lambda: (running_mean, running_var))
    x = (x - mean) / tf.sqrt(var + 1e-6)
    if learnable:
        x = x * tf.sqrt(learnable_var + 1e-6) + learnable_mean

    return x

def layer_norm(x, K=1, learnable=True, name='ln'):
    '''Apply layer norm to input x, i.e., turn output of each sample into a (learnable) gaussian distribution. 
    This function will introduce new in-trainable variables into models. 
    Its Useful to insert this function between layers to get magical improvements/degradation.

    Parameters
    ------
        x, tensor
            input time series of shape [batch_size, spatial_dim_1, ..., spatial_dim_S, feature_dim]
        K, int, default 1
            along last K dimensions to average on
        learnable, bool, default True
            whether to introduce learnable mean and var to target distribution. Becareful that layer norm use 
            element-wise learnable parameter (different from batch norm that use scalar), which may result in
            very large parameter size.
        name, str, default "bn"
            the name prefix for new variables
    Returns
    ------
        tensor
            same shape with x
    '''
    num_dims = x.shape.ndims
    N = list(map(lambda _x: _x.value, x.shape))

    if learnable:
        learnable_mean = tf.get_variable(name+'_learnable_mean', N[-K:], initializer=tf.constant_initializer(0.0))
        learnable_var = tf.get_variable(name+'_learnable_var', N[-K:], initializer=tf.constant_initializer(1.0))
    mean, var = tf.nn.moments(x, list(range(num_dims))[-K:], keep_dims=True)
    x = (x - mean) / tf.sqrt(var + 1e-6)
    if learnable:
        x = x * tf.sqrt(learnable_var + 1e-6) + learnable_mean

    return x

def _unzip_list(nested):
    '''unzip a list generated by TFModel._zip_run.
    Example: [[a1,a2,a3],[b1,b2,b3]] -> [[a1,b1],[a_2,b_2],[a3,b_3]]
    the basic elements should not be empty list [].
    '''
    # detect depth
    depth = 0
    _temp = nested
    while 1:
        if type(_temp) is list:
            depth = depth + 1
            _temp = _temp[0]
        else:
            break
    
    # unzip nested
    if depth==1:
        nested = [nested]
    elif depth>2:
        for _ in range(depth-2):
            nested = list(itertools.chain(*nested))
    nested = list(map(list, zip(*nested)))
    return nested

def stacked_window(x, width, shift=0):
    '''Compute moving stacked window on a given time series.

    Parameters
    ------
        x, tensor
            input time series of shape [batch_size, time_length, feature_dim]
        width, int
            window width
        shift, int, default 0
            decide the index of given time stamp in the output.
            For example, for x = [1,2,3,4], width = 3, and
                shift = 0, output will be [[1,2,3], [2,3,4], [3,4,0], [4,0,0]];
                shift = 1, output will be [[0,1,2], [1,2,3], [2,3,4], [3,4,0]];
                shift = -1,output will be [[2,3,4], [3,4,0], [4,0,0], [0,0,0]].

    Returns
    ------
        tensor
            output stacked time series of shape [batch_size, time_length, width, feature_dim]
    '''
    N = x.shape[0].value
    T = x.shape[1].value
    D = x.shape[2].value
    _x = tf.pad(x, [[0, 0], [max(shift, 0), max(width - 1 - shift, 0)], [0, 0]])
    _T = _x.shape[1].value
    _x = tf.transpose(_x, [0, 2, 1])
    _x = tf.reshape(_x, [N*D, _T, 1])
    w = tf.constant(np.eye(width).reshape([width, 1, width]), dtype=x.dtype)
    result = tf.nn.conv1d(_x, w, 1, 'VALID')
    result = result[:, max(-shift, 0): max(-shift, 0) + T, :]
    result = tf.reshape(result, [N, D, T, width])
    result = tf.transpose(result, [0, 2, 3, 1])
    return result

def pearson_corr(x, y, axis=-1, keepdims=False):
    ''' Compute the pearson correlation coefficient.

    Parameters
    ------
        x, tensor
            the first input tensor
        y, tensor
            the second input tensor
        axis, int, default -1
            along which index to compute correlation
        keepdims, bool, default False
            if keep corresponding dim.
    Returns
    ------
        tensor
            output tensor
    '''
    x_mean = tf.reduce_mean(x, axis=axis, keepdims=True)
    y_mean = tf.reduce_mean(y, axis=axis, keepdims=True)
    x_norm = tf.sqrt(tf.reduce_sum((x-x_mean)**2, axis=axis, keepdims=True) + 1e-8)
    y_norm = tf.sqrt(tf.reduce_sum((y-y_mean)**2, axis=axis, keepdims=True) + 1e-8)
    _x = (x - x_mean) / x_norm
    _y = (y - y_mean) / y_norm
    corr = tf.reduce_sum(_x*_y, axis=axis, keepdims=keepdims)

    return corr

def percentile(x, percentile=50, axis=-1, keepdims=False):
    ''' Compute the percentile number. return a nearest member at target percentile position.

    Parameters
    ------
        x, tensor
            the first input tensor
        percentile, int, default 50
            the percentile number between 0 and 100
        axis, int, default -1
            along which dim to compute
        keepdims, bool, default False
            if keep corresponding dim.
    Returns
    ------
        tensor
            output tensor
    '''
    shape = list(map(int, x.shape))
    axis = axis % len(shape)
    med_num = round((shape[axis] - 1) * (percentile/100.0))
    _x = tf.transpose(x, [i for i in range(len(shape)) if i != axis] + [axis])
    med = tf.nn.top_k(_x, k=shape[axis]-med_num).values[..., -1]
    if keepdims:
        med = tf.expand_dims(med, axis)
    
    return med

def median(x, axis=-1, keepdims=False):
    ''' Compute the median number. different from percentile, this function will return average of middle two elements if total number is even.

    Parameters
    ------
        x, tensor
            the first input tensor
        axis, int, default -1
            along which dim to compute
        keepdims, bool, default False
            if keep corresponding dim.
    Returns
    ------
        tensor
            output tensor
    '''
    shape = list(map(int, x.shape))
    axis = axis % len(shape)
    med_num = shape[axis]//2+1
    _x = tf.transpose(x, [i for i in range(len(shape)) if i != axis] + [axis])
    med = tf.nn.top_k(_x, k=med_num).values
    if shape[axis] % 2 == 0:
        med = (med[..., -1] + med[..., -2]) / 2
    else:
        med = med[..., -1]
    if keepdims:
        med = tf.expand_dims(med, axis)
    
    return med

def KL_diagonal_gaussian(mu1, sigma1, mu2, sigma2, axis=-1, keepdims=False):
    ''' [tensorflow] Compute the KL-Divergence between two diagonal gaussian:

    KL[p1||p2] = 1/2 * {log(det(sigma2)/det(sigma1)) - d + tr(sigma2^(-1)sigma1) + (mu2-mu1)'sigma2^(-1)(mu2-mu1)}

    Parameters
    ------
        mu1, mu2, tensor
            the mean vector of gaussian, mu2 and sigma2 can be given as scalar constant (which will be boardcasted)
        sigma1, sigma2, tensor
            the diagonal sigma of gaussian, not covariance (which is considered to be sigma**2)
        axis , int, default -1
            along which axis to operate
        keepdims, bool, default False
            if keep corresponding dim.
    Returns
    ------
        tensor
            output tensor
    '''
    
    if type(mu2) is int or type(mu2) is float:
        mu2 = tf.zeros_like(mu1) + mu2
    if type(sigma2) is int or type(sigma2) is float:
        sigma2 = tf.zeros_like(sigma1) + sigma2

    shape = list(map(int, mu1.shape))
    axis = axis % len(shape)
    term1 = tf.reduce_sum(2*tf.log(sigma2), axis=axis, keepdims=keepdims) - tf.reduce_sum(2*tf.log(sigma1), axis=axis, keepdims=keepdims)
    term2 = shape[axis]
    term3 = tf.reduce_sum((sigma1/sigma2)**2, axis=axis, keepdims=keepdims)
    term4 = tf.reduce_sum((mu2 - mu1)**2/(sigma2**2), axis=axis, keepdims=keepdims)
    kl = 1/2 * (term1 - term2 + term3 + term4)

    return kl



import numpy as np

def _gauss_ll(x, mean, sigma):
    D = x.shape[-1]
    sigma = sigma + np.abs(mean) * 1e-8
    ll = - 1/2*np.sum((x-mean)**2/sigma**2, -1) - np.sum(np.log(sigma), -1) - 1/2*D*np.log(2*np.pi)
    return ll

def kde_gauss_density_estimator(support, target, sigma):
    '''Using Gaussian kernel to estimate the mean likelihood of N target samples, on the distribution where K support samples is from.

    Parameter
    -----
        support: ndarry, shape (..., K, D)
        target: ndarry, shape (..., N, D)
        sigma: ndarry, shape (..., D)

    Return
    -----
        ndarry, shape (...)
    '''
    _support = support[..., None, :, :]
    _target = target[..., :, None, :]
    _sigma = sigma[..., None, None, :]
    ll = _gauss_ll(_target, _support, _sigma) # (..., N, K)
    ll_max = np.max(ll, -1, keepdims=True)
    estimate_ll = np.log(np.mean(np.exp(ll - ll_max), -1, keepdims=True)) + ll_max
    mean_ll = np.mean(estimate_ll, (-1, -2))
    return mean_ll

def kde_find_diagonal_sigma(support, folds=5, min=0.1, max=1.0, n=20):
    '''Use K-folds cross-validation to find best diagonal gauss kernel width. (one sigma for all distributions in a batch)

    Parameter
    -----
        support: ndarry
            input support samples, shape (..., K, D)
        folds: int, default 5
            number of k-folds cross-validation
        min: float, default 0.1
            start of search interval (include)
        max: float, default 1.0
            end of search interval (include)
        n: int, default 20
            number of points included in search interval

    Return
    -----
        ndarry
            shape (D,)
    '''
    N, D = support.shape[-2:]
    N = N//folds*folds
    _support = support[..., :N, :]
    _support_split = np.split(_support, folds, -2)
    search_range = np.linspace(min, max, n)

    best_sigma = []
    for i in range(D):
        val_result = []
        for k in range(folds):
            _support_train = np.concatenate([_support_split[j][..., i:i+1] for j in range(folds) if j!=k], -2)
            _support_val = _support_split[k][..., i:i+1]
            _val_result = [np.mean(kde_gauss_density_estimator(_support_train, _support_val, np.array([h]))) for h in search_range]
            val_result.append(_val_result)
        val_result = np.mean(val_result, 0)
        best_sigma.append(search_range[np.argmax(val_result)])

    return np.array(best_sigma)

def kde_find_identity_sigma(support, folds=5, min=0.1, max=1.0, n=20):
    '''Use K-folds cross-validation to find best identity gauss kernel width. (one sigma for all distributions in a batch)

    Parameter
    -----
        support: ndarry
            input support samples, shape (..., K, D)
        folds: int, default 5
            number of k-folds cross-validation
        min: float, default 0.1
            start of search interval (include)
        max: float, default 1.0
            end of search interval (include)
        n: int, default 20
            number of points included in search interval

    Return
    -----
        ndarry
            shape (1,)
    '''
    N, D = support.shape[-2:]
    N = N//folds*folds
    _support = support[..., :N, :]
    _support_split = np.split(_support, folds, -2)
    search_range = np.linspace(min, max, n)

    val_result = []
    for k in range(folds):
        _support_train = np.concatenate([_support_split[j] for j in range(folds) if j!=k], -2)
        _support_val = _support_split[k]
        _val_result = [np.mean(kde_gauss_density_estimator(_support_train, _support_val, np.array([h]))) for h in search_range]
        val_result.append(_val_result)
    val_result = np.mean(val_result, 0)
    best_sigma = search_range[np.argmax(val_result)]

    return np.array([best_sigma])

def kde_find_rule_of_thumb_sigma(support):
    '''Use rule-of-thumb (which use gaussian assumption) to find best diagonal gauss kernel width. This provide different estimation for different distribution.

    Parameter
    -----
        support: ndarry, shape (..., K, D)

    Return
    -----
        ndarry
            shape (..., D)
    '''
    K = support.shape[-2]
    std = np.std(support, -2)
    best_sigma = 1.06 * std * np.power(K, -0.2)
    return best_sigma



# -*- coding: UTF-8 -*-
import numpy as np
import pandas as pd
import time
import json
import os
import requests

from apdt.config import _config

class DarkSkyAPI():
    '''DarkSkyAPI is a spider on darsky.net. Help you download weather history data for a set of sites using a set of accounts.

    darksky_secret_key_num and darksky_secret_key_0 should be decleared in config.py.

    Method
    ------
        __init__: Initialize.
            sites: a list of sites to collect data on. e.g. ['1001A', '1414A'].
            start_date: the start date (include) to collect data. e.g. '2019-01-01'.
            end_date: the end date (include) to collect data. e.g. '2019-12-31'.
            DailyAccountQuota: the maximum call quota for each accounts per day. e.g. (default) 990.
            save_path: the path to put download data. e.g. (default) 'data/weather/'.

        run: excute download.
            flag: set True to start download. default: False.
            shift: claim a start point for task, this can be useful for continue a job from previous start. default: 0.

    '''
    def __init__(self, sites, start_date, end_date, **kwarg):
        if type(sites) is str:
            sites = [sites]
        if 'DailyAccountQuota' not in kwarg.keys():
            kwarg['DailyAccountQuota'] = 990
        if type(kwarg['DailyAccountQuota']) is not list:
            kwarg['DailyAccountQuota'] = [kwarg['DailyAccountQuota'] for _ in range(_config['darksky_secret_key_num'])]
        if 'save_path' not in kwarg.keys():
            kwarg['save_path'] = 'data/weather/'

        self.site_list = pd.read_csv(_config['nms_data_path'] + 'site_location.csv')
        self.sites = sites
        self.start_date = start_date
        self.end_date = end_date
        self.quota = kwarg['DailyAccountQuota']
        self.save_path = kwarg['save_path']
        self.req_pool = self._requests_pool()
        self.account_num = _config['darksky_secret_key_num']
        self.secret_key_list = [_config['darksky_secret_key_' + str(i)] for i in range(self.account_num)]
        self.entry_num = len(self.req_pool)
        self.est_days = self.entry_num // sum(kwarg['DailyAccountQuota']) + 1
        ...

    def _requests_pool(self):
        req_pool = []
        for s in self.sites:
            start_time = time.mktime((time.strptime(self.start_date+' 12:00:00','%Y-%m-%d %H:%M:%S')))
            end_time = time.mktime((time.strptime(self.end_date+' 12:00:00','%Y-%m-%d %H:%M:%S')))
            lon = self.site_list.loc[self.site_list['监测点编码'] == s, '经度'].values[0]
            lat = self.site_list.loc[self.site_list['监测点编码'] == s, '纬度'].values[0]
            req_loc = '{0:f},{1:f},'.format(lat, lon)
            time_now = start_time
            while time_now <= end_time:
                req_0 = 'https://api.darksky.net/forecast/'
                req_1 = '/'+req_loc+str(round(time_now))
                file = self.save_path + s + '/' + time.strftime('%Y-%m-%d', time.gmtime(time_now)) + '.json'
                req_pool.append([req_0, req_1, file])
                time_now += 3600*24
        return req_pool

    def run(self, flag=False, shift=0, wait_hour=24):
        est_days = (self.entry_num - shift) // sum(self.quota) + 1
        print('This task include', self.entry_num - shift, ' entries and need ', est_days, ' days using ', self.account_num, ' accounts.')
        if flag:
            key = 0
            key_count = 0
            for s in self.sites:
                if not os.path.exists(self.save_path + s + '/'):
                    os.mkdir(self.save_path + s + '/')
            for i in range(shift, self.entry_num):
                if (i+1)%500 == 0:
                    print(i+1, '/', self.entry_num, '...')
                if key_count >= self.quota[key]:
                    # print('key ', key, ' run out of quota')
                    key_count = 0
                    key = key + 1
                    if key >= self.account_num:
                        key = 0
                        print('Waiting', wait_hour,' hours...')
                        print('Current shift:', i)
                        time.sleep(wait_hour*3600)
                req = self.req_pool[i]
                url = req[0] + self.secret_key_list[key] + req[1]
                file = req[2]
                try:
                    r = requests.get(url)
                    with open(file, 'w') as result_file:
                        result_file.write(r.text)
                except:
                    print('Error: ', file)
                    with open('ERROR_LOG_FILE', 'a') as ERROR_LOG_file:
                        ERROR_LOG_file.write(url+'\n')
                        ERROR_LOG_file.write(file+'\n')
                key_count = key_count + 1
        else:
            print('Use DarkSkyAPI.run(flag = True) to confirm and run task.')

    def retry_error(self, flag=False, shift=0, wait_hour=24):
        with open('ERROR_LOG_FILE', 'r') as ERROR_LOG_file:
            lines = [x.replace('\n','') for x in ERROR_LOG_file.readlines()]
            req_pool = [['https://api.darksky.net/forecast/', lines[2*i].rsplit('/',2)[-1], lines[2*i+1]] for i in range(len(lines)//2)]

        with open('ERROR_LOG_FILE_BCK', 'w') as ERROR_LOG_file:
            for line in lines:
                ERROR_LOG_file.write(line+'\n')
        with open('ERROR_LOG_FILE', 'w') as ERROR_LOG_file:
            ...
        
        entry_num = len(req_pool)
        est_days = (entry_num - shift) // sum(self.quota) + 1
        print('This task include', entry_num - shift, ' entries and need ', est_days, ' days using ', self.account_num, ' accounts.')
        
        if flag:
            key = 0
            key_count = 0
            for s in self.sites:
                if not os.path.exists(self.save_path + s + '/'):
                    os.mkdir(self.save_path + s + '/')
            for i in range(shift, entry_num):
                if (i+1)%500 == 0:
                    print(i+1, '/', entry_num, '...')
                if key_count >= self.quota[key]:
                    key_count = 0
                    key = key + 1
                    if key >= self.account_num:
                        key = 0
                        print('Waiting', wait_hour,' hours...')
                        print('Current shift:', i)
                        time.sleep(wait_hour*3600)
                req = req_pool[i]
                url = req[0] + self.secret_key_list[key] + '/' + req[1]
                file = req[2]
                try:
                    r = requests.get(url)
                    with open(file, 'w') as result_file:
                        result_file.write(r.text)
                except:
                    print('Error: ', file)
                    with open('ERROR_LOG_FILE', 'a') as ERROR_LOG_file:
                        ERROR_LOG_file.write(url+'\n')
                        ERROR_LOG_file.write(file+'\n')
                key_count = key_count + 1
        else:
            print('Use DarkSkyAPI.retry_error(flag = True) to confirm and run task.')


# pylint: disable=E1101
import numpy as np
import pandas as pd
import os
import time
import apdt
import pickle
import scipy.stats
from apdt.ml.utils import _unzip_list

try:
    import tensorflow as tf
except:
    pass

class DataSet():
    '''DataSet can make DataPack object being Machine Learning ready.
    Description
    -----------
        Dataset object consist of three parts: pre_process, get_data, and post_process.

        - pre_process is to make DataPack obejct into a ML-ready ndarray self.tr, self.val and self.te, which have size N*D_1*D_2*...; The first dimension is considered as independent sample index; Beside this, three variables self.tr_batch_num, self.val_batch_num and self.te_batch_num which claims the coressponding sample number should also be defined.
        - get_data has three key methods tr_get_batch, val_get_batch and te_get_batch, will return a B*D_1*D_2*... as a batch of training/validation/testing data, or list of ndarray when you need multi-input.
        - post_process defined the action that after every get_data call.

        In most case users have to re-define pre_process function to make their desired dataset structure. We also provide some pre-defined useful pre_process function for you. get_data and post_process typically don't need much work, but they can still be re-defined if you like.

    Parameters
    ----------
        datapack: DataPack object or str
            The datapack to be processed into a dataset. If a str is given, it will be considered as a pikle file path, and will construct dataset directly from that file.
        method: str
            If this argument is defined, we will use sepecific pre-defined pre_process fuction.
            Avaliable:
            - 'window': use a sliding window across time-axis to generate samples. locations in a time slice are supposed to be same (i.e. fixed-station). a samlple will be a N*T*D array. where N is the number of locations, T is window length and D is feature dimensions.
                argument for 'window':
                - split_ratio=0.7: the ratio of training set. Example: when split_ratio=0.7, training set will take first 70% data, validation set and testing set will be identical and take last 30%. when split_ration=(0.5, 0.25, 0.25), training set will take first 50% data, validation set will take middle 25% data, and testing set take last 25% data. When given None, there's no splition, training/validation/testing will use same entire dataset.
                - seq_len=100: the window length;
                - strides=seq_len-1: the sliding step;
                - sub_sample=1: the sub_sample rate. The data time span should be at least seq_len*sub_sample (instead of (seq_len-1)*sub_sample+1) to ensure one sample
                - normalize=True: if we do 0-1 normalize, Warning: dataset should always be properly normalized, and this operation is a in-place operation.
                - feature=['data0', 'data1', ...]: the feature list to be used. default to all features avaliable.
        seed: int
            The initial random seed, default no seed.
        shuffle_train_only: bool, default False
            if False, Train/Val/Test set will be shuffled at every glance. if True, only Train set will.

    '''
    def __init__(self, datapack, **kwarg):
        if 'supervised' not in kwarg.keys():
            kwarg['supervised'] = False
        if 'seed' not in kwarg.keys():
            kwarg['seed'] = None
        if 'shuffle_train_only' not in kwarg.keys():
            kwarg['shuffle_train_only'] = False

        self._kwarg = kwarg
        self.supervised = kwarg['supervised']

        if type(datapack) is apdt.DataPack:
            self.pre_process(datapack, **kwarg)
        elif type(datapack) is str:
            if datapack=='MNIST':
                self.MNIST(**kwarg)
            else:
                self.load(datapack, **kwarg)

        self._init_counter(**kwarg)

    def dump(self, path):
        dumped_data = {'data': self.data, 'tr': self.tr, 'val': self.val, 'te': self.te}
        with open(path, 'wb') as f:
            pickle.dump(dumped_data, f, 1)

    def load(self, path, **kwarg):
        with open(path, 'rb') as f:
            dumped_data = pickle.load(f)
        self.data = dumped_data['data']
        self.tr = dumped_data['tr']
        if 'val' in dumped_data.keys():
            self.val = dumped_data['val']
        else:
            self.val = dumped_data['te']
            print('WARNING: ' + path + 'is generated by early version of apdt and will not be supported soon.')
        self.te = dumped_data['te']
        self.tr_batch_num = self.tr.shape[0]
        self.val_batch_num = self.val.shape[0]
        self.te_batch_num = self.te.shape[0]

    def pre_process(self, datapack, **kwarg):
        if 'method' not in kwarg.keys():
            raise Exception("You have to re-define pre_process function or use a pre-defined one using 'method' argument.")

        if kwarg['method'] == 'window':
            if 'time-aligned' not in datapack.tag:
                raise Exception("datapack should be aligned along time axis first.")
            if 'fixed-location' not in datapack.tag:
                raise Exception("window method only support fixed station data.")
            
            if 'split_ratio' not in kwarg.keys():
                kwarg['split_ratio'] = 0.7
            if 'sub_sample' not in kwarg.keys():
                kwarg['sub_sample'] = 1
            if 'seq_len' not in kwarg.keys():
                kwarg['seq_len'] = 100
            if 'strides' not in kwarg.keys():
                kwarg['strides'] = kwarg['seq_len']
            if 'normalize' not in kwarg.keys():
                kwarg['normalize'] = True
            if 'normalize_method' not in kwarg.keys():
                kwarg['normalize_method'] = '99pt'
            if 'feature' not in kwarg.keys():
                kwarg['feature'] = [x for x in datapack.data.columns if x.startswith('data')]

            self.method = 'window'
            T = len(datapack.data.index.drop_duplicates()) 
            N = len(datapack.data.site_id.drop_duplicates())
            D = len(kwarg['feature'])
            if kwarg['normalize']:
                datapack = apdt.proc.linear_normalize(datapack, kwarg['feature'], kwarg['normalize_method'])
            self.data = datapack.data.reset_index().sort_values(['datetime', 'site_id'])[kwarg['feature']].values.reshape((T, N, D))

            if kwarg['split_ratio'] is None:
                self.tr = self.data
                self.val = self.data
                self.te = self.data
            elif type(kwarg['split_ratio']) is float:
                split_point = int(T * kwarg['split_ratio'])
                self.tr = self.data[:split_point]
                self.val = self.data[split_point:]
                self.te = self.data[split_point:]
            elif type(kwarg['split_ratio']) is tuple:
                split_point_1 = int(T * kwarg['split_ratio'][0])
                split_point_2 = split_point_1 + int(T * kwarg['split_ratio'][1])
                self.tr = self.data[:split_point_1]
                self.val = self.data[split_point_1:split_point_2]
                self.te = self.data[split_point_2:]
            self.tr_batch_num = (self.tr.shape[0]//kwarg['sub_sample']-kwarg['seq_len']+kwarg['strides'])//(kwarg['strides'])*kwarg['sub_sample']
            self.val_batch_num = (self.val.shape[0]//kwarg['sub_sample']-kwarg['seq_len']+kwarg['strides'])//(kwarg['strides'])*kwarg['sub_sample']
            self.te_batch_num = (self.te.shape[0]//kwarg['sub_sample']-kwarg['seq_len']+kwarg['strides'])//(kwarg['strides'])*kwarg['sub_sample']
            if self.tr_batch_num <= 0 or self.val_batch_num <= 0 or self.te_batch_num <= 0:
                raise Exception("time_length is not enough to construct a window.")
            
            self.tr_list = []
            self.val_list = []
            self.te_list = []
            for i in range(self.tr_batch_num//kwarg['sub_sample']):
                for j in range(kwarg['sub_sample']):
                    self.tr_list.append(self.tr[i*(kwarg['strides']*kwarg['sub_sample'])+j:i*(kwarg['strides']*kwarg['sub_sample'])+kwarg['seq_len']*kwarg['sub_sample']:kwarg['sub_sample']])
            for i in range(self.val_batch_num//kwarg['sub_sample']):
                for j in range(kwarg['sub_sample']):
                    self.val_list.append(self.val[i*(kwarg['strides']*kwarg['sub_sample'])+j:i*(kwarg['strides']*kwarg['sub_sample'])+kwarg['seq_len']*kwarg['sub_sample']:kwarg['sub_sample']])
            for i in range(self.te_batch_num//kwarg['sub_sample']):
                for j in range(kwarg['sub_sample']):
                    self.te_list.append(self.te[i*(kwarg['strides']*kwarg['sub_sample'])+j:i*(kwarg['strides']*kwarg['sub_sample'])+kwarg['seq_len']*kwarg['sub_sample']:kwarg['sub_sample']])
            self.tr = np.array(self.tr_list)
            self.val = np.array(self.val_list)
            self.te = np.array(self.te_list)
            self.tr = np.transpose(self.tr, (0,2,1,3))
            self.val = np.transpose(self.val, (0,2,1,3))
            self.te = np.transpose(self.te, (0,2,1,3))

    def _init_counter(self, **kwarg):
        if kwarg['seed'] is not None:
            np.random.seed(kwarg['seed'])
        self.tr_batch_counter = 0
        self.val_batch_counter = 0
        self.te_batch_counter = 0
        self.tr_batch_perm = np.linspace(0, self.tr_batch_num-1, self.tr_batch_num, dtype=np.int32)
        self.val_batch_perm = np.linspace(0, self.val_batch_num-1, self.val_batch_num, dtype=np.int32)
        self.te_batch_perm = np.linspace(0, self.te_batch_num-1, self.te_batch_num, dtype=np.int32)
        np.random.shuffle(self.tr_batch_perm)
        if not kwarg['shuffle_train_only']:
            np.random.shuffle(self.val_batch_perm)
            np.random.shuffle(self.te_batch_perm)

    def tr_get_batch(self, batch_size=1):
        if batch_size > self.tr_batch_num:
            batch_size = self.tr_batch_num
        if self.tr_batch_counter + batch_size > self.tr_batch_num:
            np.random.shuffle(self.tr_batch_perm)
            self.tr_batch_counter = 0
        target_index = self.tr_batch_perm[self.tr_batch_counter: self.tr_batch_counter + batch_size]
        self.tr_batch_counter = self.tr_batch_counter + batch_size
        if self.supervised:
            batchx = self.tr[target_index]
            batchy = self.tr_y[target_index]
            batch = self.post_process([batchx, batchy], mode='train')
        else:
            batch = self.tr[target_index]
            batch = self.post_process(batch, mode='train')
        return batch

    def val_get_batch(self, batch_size=1):
        if batch_size > self.val_batch_num:
            batch_size = self.val_batch_num
        if self.val_batch_counter + batch_size > self.val_batch_num:
            if not self._kwarg['shuffle_train_only']:
                np.random.shuffle(self.val_batch_perm)
            self.val_batch_counter = 0
        target_index = self.val_batch_perm[self.val_batch_counter: self.val_batch_counter + batch_size]
        self.val_batch_counter = self.val_batch_counter + batch_size
        if self.supervised:
            batchx = self.val[target_index]
            batchy = self.val_y[target_index]
            batch = self.post_process([batchx, batchy], mode='validate')
        else:
            batch = self.val[target_index]
            batch = self.post_process(batch, mode='validate')
        return batch
    
    def te_get_batch(self, batch_size=1):
        if batch_size > self.te_batch_num:
            batch_size = self.te_batch_num
        if self.te_batch_counter + batch_size > self.te_batch_num:
            if not self._kwarg['shuffle_train_only']:
                np.random.shuffle(self.te_batch_perm)
            self.te_batch_counter = 0
        target_index = self.te_batch_perm[self.te_batch_counter: self.te_batch_counter + batch_size]
        self.te_batch_counter = self.te_batch_counter + batch_size
        if self.supervised:
            batchx = self.te[target_index]
            batchy = self.te_y[target_index]
            batch = self.post_process([batchx, batchy], mode='test')
        else:
            batch = self.te[target_index]
            batch = self.post_process(batch, mode='test')
        return batch

    def post_process(self, batch, **kwarg):
        '''Define the action on post_process of batches from dataset, default doing nothing.

        Parameters
        ------
            batch, tensor
            mode, in ['train, 'validate', 'test]
        '''
        return batch

    def MNIST(self, **kwarg):
        # Load data from keras interface
        import tensorflow as tf
        mnist = tf.keras.datasets.mnist
        (x_train, y_train), (x_test, y_test) = mnist.load_data()
        x_train, x_test = x_train / 255.0, x_test / 255.0
        y_train = (np.outer(np.ones((y_train.shape[0],)), np.arange(10)) == y_train.reshape([-1, 1])) * 1.0
        y_test = (np.outer(np.ones((y_test.shape[0],)), np.arange(10)) == y_test.reshape([-1, 1])) * 1.0

        # Construct dataset
        self.supervised = True
        self.tr = x_train
        self.val = x_test
        self.te = x_test
        self.tr_y = y_train
        self.val_y = y_test
        self.te_y = y_test
        self.tr_batch_num = self.tr.shape[0]
        self.val_batch_num = self.val.shape[0]
        self.te_batch_num = self.te.shape[0]

class TFModel():
    '''Provide model constructure interface based on Tensorflow (1.x).
    Description
    -----------
        There are some method of this class:

        - def_model(): you HAVE TO re-define this function to build your deep network. this fuction should define:
            - self.input, self.input should be a placeholder tensor or a list of such, where training batch passed to TFModel will be sent into;
            - self.metric, self.metric should be a list of tensor, which will be evaluated and printed during training, and will by default be set to [self.loss];
            - self.loss, self.loss should be a tensor of single float value, which will be minimized by optimizer; (after 0.1.8) self.loss can be a list to define complex optimization like GAN.
            - self.pred. self.pred should be a tensor or a list of such, which will be the output to be evaluated given some input.
            - self.learning_rate (optional), self.learning_rate should be a placeholder tensor of single float value, which will control the gradient scale;
            - self.pretrain_var_map (optional), self.pretrain_var_map is only needed when pre-train model is given. The map between model variables, which need pre-train values instead of random initialization, to its name in pretrain model. By default {v.op.name: v for v in tf.trainable_variables()}.

        - fit(dataset): fit the model using given DataSet object. Note that please make sure the dataset structure is compatible with model's input interface.

        - eval(data): evaluate the model using given data (ndarray or a list of such, depends on self.input).

        - load(path): load parameters at given path.

    Parameters
    ----------
        - model.__init__()
            - l2_norm, float, default None.
                A L2-Regularization will apllied with weight decay parameter as l2_norm, It will be disabled when given None.
                l2_norm should not be used together with multiple loss.
            - clip_gvs, bool, default False.
                all gradients in network will be cliiped to [-1,1] when set True.
            - seed, int, default to currrent time.
                The random seed to be set in Numpy and Tensorflow.
            - pretrain_model, str, default None.
                The pretrain model path.
            - optimizer, str, default adam.
                available: 'adam', 'rmsprop'
        
        - model.fit():
            - model_name='NewModel', str.
            - lr=1e-3, float.
            - baseline=0, float. if the model loss is smaller than this, it will be automatically saved.
            - epoch=100, int. How many times should we go through the dataset.
            - print_every_n_epochs=1. int.
            - test_every_n_epochs=1. int.
            - lr_annealing='constant'. str. How to decay learning rate during training. optional: 'constant', 'step', 'cosine'.
            - lr_annealing_step_length=epoch/4. int. Avaliable when lr_annealing is set to 'step'. How often do we decay the learning rate.
            - lr_annealing_step_divisor=2.0. float. Avaliable when lr_annealing is set to 'step', we will apply lr=lr/lr_annealing_step_divisor to slow down the training process.
            - early_stop=None. int. Stop training if there is no better validation result since last recorder. This parameter decides the waiting epoch number. If None this feature will be disabled.
            - repeat=1. int. If set >1, we will randomly reset parameters and retrain the model for certain times, and compute statistical performance.
            - rstat='pred'. str. The type of reported repeat statistics. 'pred': gaussian 95% predictive interval. 't': student-t 95% confidence interval
            - verbose=0. int. The verbose level. 0-print all. 1-only summary. 2-mute.
            - unzip="cartesian". str. decide how to treat list input from dataset. optional: "cartesian" (return a cartesian product of all list), "parallel" (combine list members in order)
    '''
    def __init__(self, **kwarg):
        # Parameter Check
        if 'seed' not in kwarg.keys():
            kwarg['seed'] = None
        if 'pretrain_model' not in kwarg.keys():
            kwarg['pretrain_model'] = None
        self.kwarg = kwarg
        
        # Initial tensor
        self.training = tf.placeholder(tf.bool)
        self.learning_rate = tf.placeholder(tf.float32)
        self.training_process = tf.placeholder(tf.float32)
        self.input = [None]
        self.pred = None
        self.loss = None
        self.metric = None
        self.metric_name = None
        self.pretrain_var_map = None
        self.summary_merged = None
        self.var_list = None

        # Define Model
        self.def_model(**kwarg)
        if self.loss is not None:
            if type(self.loss) is not list:
                self.loss = [self.loss]
            if self.var_list is None:
                self.var_list = [tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES) for _ in range(len(self.loss))]
            if self.metric is None:
                self.metric = self.loss
            elif type(self.metric) is not list:
                self.metric = [self.metric]
            if self.metric_name is None:
                self.metric_name = ['metric_' + str(i) for i in range(len(self.metric))]
            if len(self.metric_name) < len(self.metric):
                _diff = len(self.metric) - len(self.metric_name)
                self.metric_name = self.metric_name + ['metric_' + str(i) for i in range(_diff)]
            self.setup_train_op(**kwarg)
        else:
            raise Exception('self.loss not defined in your model.')        
        if self.pred is None:
            self.pred = self.metric[0]
        if self.summary_merged is None and len(tf.get_collection(tf.GraphKeys.SUMMARIES)) > 0:
            self.summary_merged = tf.summary.merge_all()
        self.variables_num = np.sum([np.prod(v.get_shape().as_list()) for v in tf.trainable_variables()])
        
        # Start Engine
        self.saver = tf.train.Saver(max_to_keep=None)
        if kwarg['pretrain_model'] is not None:
            if type(kwarg['pretrain_model']) is str:
                if self.pretrain_var_map is None:
                    self.pretrain_var_map = {v.op.name: v for v in tf.trainable_variables()}
                self._pre_saver = tf.train.Saver(var_list=self.pretrain_var_map, max_to_keep=None)
            elif type(kwarg['pretrain_model']) is list:
                if self.pretrain_var_map is None:
                    self.pretrain_var_map = [{v.op.name: v for v in tf.trainable_variables()} for _ in range(len(kwarg['pretrain_model']))]
                self._pre_saver = [tf.train.Saver(var_list=x, max_to_keep=None) for x in self.pretrain_var_map]
            else:
                raise Exception("pretrain_model must be a path or a list of path.")
        self.sess = tf.Session()
        self.init_parameter()

    def init_parameter(self, seed=None):
        # Parameter Initialize
        if seed is None:
            if self.kwarg['seed'] is not None:
                seed = self.kwarg['seed']
            else:
                seed = int(time.time())
        np.random.seed(seed)
        tf.set_random_seed(seed)
        self.sess.run(tf.global_variables_initializer())
        if self.kwarg['pretrain_model'] is not None:
            if type(self.kwarg['pretrain_model']) is str:
                self._pre_saver.restore(self.sess, save_path=self.kwarg['pretrain_model'])
            elif type(self.kwarg['pretrain_model']) is list:
                for i in range(len(self.kwarg['pretrain_model'])):
                    self._pre_saver[i].restore(self.sess, save_path=self.kwarg['pretrain_model'][i])

    def def_model(self, **kwarg):
        pass

    def setup_train_op(self, **kwarg):
        # Parameter check
        if 'l2_norm' not in kwarg.keys():
            kwarg['l2_norm'] = None
        if 'keep_norm_loss' not in kwarg.keys():
            kwarg['keep_norm_loss'] = True
        if 'clip_gvs' not in kwarg.keys():
            kwarg['clip_gvs'] = False
        if 'optimizer' not in kwarg.keys():
            kwarg['optimizer'] = "adam"

        # Set up optimizer
        if kwarg['optimizer']=="adam":
            optimizer = tf.train.AdamOptimizer(learning_rate=self.learning_rate)
        elif kwarg['optimizer']=="rmsprop":
            optimizer = tf.train.RMSPropOptimizer(learning_rate=self.learning_rate)
        elif kwarg['optimizer']=="sgd":
            optimizer = tf.train.GradientDescentOptimizer(learning_rate=self.learning_rate)
        
        # Compute Gradient
        gvs = []
        for i in range(len(self.loss)):
            # Add global regulizer
            if kwarg['l2_norm']:
                l2_loss =  kwarg['l2_norm'] * tf.add_n([tf.nn.l2_loss(tf.cast(v, tf.float32)) for v in tf.trainable_variables()])
                if kwarg['keep_norm_loss']:
                    self.loss[i] = self.loss[i] + l2_loss
                    _gvs = optimizer.compute_gradients(self.loss[i], var_list=self.var_list[i])
                else:
                    _gvs = optimizer.compute_gradients(self.loss[i] + l2_loss, var_list=self.var_list[i])
            else:
                _gvs = optimizer.compute_gradients(self.loss[i], var_list=self.var_list[i])

            # Clip Gradients
            if kwarg['clip_gvs']:
                clipped_gvs = [(tf.clip_by_value(grad, -1., 1.), var) for grad, var in _gvs if grad is not None]
                _gvs = [clipped_gvs[i] for i in range(0,len(clipped_gvs))]
            
            gvs.append(_gvs)

        # Check update ops.
        update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
        with tf.control_dependencies(update_ops):
                self.train_op = [optimizer.apply_gradients(gvs[0])]
        for i in range(1, len(self.loss)):
            self.train_op.append(optimizer.apply_gradients(gvs[i]))
        if len(self.loss)>1 and len(update_ops)>0:
            print("WARNING: Update OP is bounded to loss[0].")

    def load(self, path):
        self.saver.restore(self.sess, save_path=path)

    def update(self, feed_dict, epoch, unzip):
        '''Apply an update on network based on given feed_dict and return statistics. This function optimize the first given loss (self.loss[0]). In case of multiple loss defined, you may reload this function for customized optimization procedure.
        Parameters
        ----------
            feed_dict, dict
                a dictionary that maps all necessary input tensor to their input value 
            epoch, int
                the epoch number of current update, useful when use dynamic update algorithm.
            unzip, str
                how to treat list input from dataset.
        
        Returns
        ----------
            out, a list of list, where: 
                out[0] is a list of real value, considered as loss in report;
                out[1+k] is a list of real value, considered as metric[k] in report;
                out[1+K] is a list of summary result, if self.summary_merged is not None;
        '''
        target = [self.train_op[0], self.loss[0]] + self.metric
        if self.summary_merged is not None:
            target = target + [self.summary_merged]
        _re = self._zip_run(target, feed_dict, mode=unzip)
        _re = _unzip_list(_re)
        return _re[1:]

    def eval(self, data):
        if type(self.input) is list:
            feed_dict = {self.input[i]: data[i] for i in range(len(self.input))}
            feed_dict.update({self.learning_rate: 0.0, self.training: False, self.training_process: 0.0})
        else:
            feed_dict = {self.learning_rate: 0.0, self.training: False, self.training_process: 0.0, self.input: data}
        result = self.sess.run(self.pred, feed_dict)
        return result

    def _zip_run(self, target, feed_dict, mode='cartesian', _r=0):
        '''Recurrently unzip lists in feed_dict to generate samples.
        Description
        -----------
            Compared to self.sess.run(target, feed_dict), self._zip_run(target, feed_dict) support list values in feed_dict, where the average of unziped feed_dict will be returned.
        
        Parameters
        ----------
            target, TF tensor or a list of TF tensor
                the tensorflow tensor to be evaluated.
            feed_dict, dict
                the feed data, list values supported.
            mode, str, default 'cartesian'
                the unzip method, optional: "cartesian" (return a cartesian product of all list), "parallel" (combine list members in order)
            _r, int
                private variable for recursive purpose, set this larger than 0 will skip first _r elements in feed_dict
        
        Returns
        ----------
            ndarray or a multi-level list of ndarray

        Examples
        ----------
            If we run self._zip_run(self.loss, {imput:[x1,x2]}) it is equivalent to do: [self.session.run(self.loss, {input: x}) for x in [x1,x2]]
        '''
        if mode=='cartesian':
            if _r == len(feed_dict.keys()):
                return self.sess.run(target, feed_dict)
            else:
                if type(list(feed_dict.values())[_r]) is list:
                    _result = []
                    for i in range(len(list(feed_dict.values())[_r])):
                        _temp_dict = feed_dict.copy()
                        _temp_dict[list(feed_dict.keys())[_r]] = list(feed_dict.values())[_r][i]
                        # change _r+1 to _r to support multi-level lists.
                        _result.append(self._zip_run(target, _temp_dict, mode, _r=_r+1))
                    return _result
                else:
                    return self._zip_run(target, feed_dict, mode, _r=_r+1)
        elif mode=='parallel':
            level = [(i, len(list(feed_dict.values())[i])) for i in range(len(feed_dict.keys())) if type(list(feed_dict.values())[i]) is list]
            if len(level)==0:
                return self.sess.run(target, feed_dict)
            else:
                _result = []
                for i in range(min([x[1] for x in level])):
                    _temp_dict = feed_dict.copy()
                    for j,_ in level:
                        _temp_dict[list(feed_dict.keys())[j]] = list(feed_dict.values())[j][i]
                    _result.append(self.sess.run(target, _temp_dict))
                return _result
        else:
            raise Exception("unexpected unzip method " + mode)

    def test(self, dataset, **kwarg):
        '''Test the model on given dataset.
        Parameters
        ------
            dataset, DataSet object.
                the dataset.
            mode='te', str.
                on which part of dataset we test on. optional: ['tr', 'val', 'te']
            target='metric', str
                which tensor we test on. optional: ['metric', 'loss']
            batch_size=1, int.
                the batch size.

        Returns
        ------
            float,
                the test result.
        '''
        if 'mode' not in kwarg.keys():
            kwarg['mode'] = 'te'
        if 'target' not in kwarg.keys():
            kwarg['target'] = 'metric'
        if 'unzip' not in kwarg.keys():
            kwarg['unzip'] = 'cartesian'
        if 'batch_size' not in kwarg.keys():
            kwarg['batch_size'] = 1
        if kwarg['batch_size'] > min(dataset.tr_batch_num, dataset.val_batch_num, dataset.te_batch_num):
            kwarg['batch_size'] = min(dataset.tr_batch_num, dataset.val_batch_num, dataset.te_batch_num)
            print('WARNING: batch_size should be smaller than dataset size. Automatically set batch_size =', kwarg['batch_size'])
        
        if kwarg['mode'] in ['tr', 'train', 'training']:
            batch_num = dataset.tr_batch_num
            get_batch = dataset.tr_get_batch
        elif kwarg['mode'] in ['val', 'validate', 'validation']:
            batch_num = dataset.val_batch_num
            get_batch = dataset.val_get_batch
        elif kwarg['mode'] in ['te', 'test', 'testing']:
            batch_num = dataset.te_batch_num
            get_batch = dataset.te_get_batch
        else:
            raise Exception('Unsupport test mode.')
        
        if kwarg['target']=='metric':
            target = self.metric
        elif kwarg['target']=='loss':
            target = self.loss
        else:
            raise Exception('Unsupport test target.')
        
        me_list = [[] for _ in range(len(target))]
        for _ in range(batch_num//kwarg['batch_size']):
            batch = get_batch(kwarg['batch_size'])
            if type(self.input) is list:
                feed_dict = {self.input[i]: batch[i] for i in range(len(self.input))}
                feed_dict.update({self.learning_rate: 0.0, self.training: False, self.training_process: 0.0})
            else:
                feed_dict = {self.learning_rate: 0.0, self.training: False, self.training_process: 0.0, self.input: batch}
            # ls = self.sess.run(self.loss, feed_dict)
            _re = self._zip_run(target, feed_dict, mode=kwarg['unzip'])
            _re = _unzip_list(_re)
            for _i in range(len(target)):
                me = np.mean(_re[_i])
                me_list[_i].append(me)
        me_list = [np.mean(x) for x in me_list]

        return me_list

    def iter_do(self, epoch):
        '''Do some customized job during training, e.g., evaluation and output. default doing nothing.
        '''
        pass

    def fit(self, dataset, **kwarg):
        '''Train the model on given dataset.

        This function will train the model on the training set, keep the best model on validation set, and report its result on testing set. This proceduremay repeat for statistical perfomence analysis.

        Parameters
        ------
            dataset, DataSet object.
                the dataset.
        Returns
        ------
            list of float.
                the final test result. list length equals to the repeat number.
            list of str.
                the model name list. list length equals to the repeat number.
        '''
        # parameter checking
        if 'model_name' not in kwarg.keys():
            kwarg['model_name'] = 'NewModel'
        if 'lr' not in kwarg.keys():
            if 'learning_rate' in kwarg.keys():
                kwarg['lr'] = kwarg['learning_rate']
            else:
                kwarg['lr'] = 1e-3
        if 'baseline' not in kwarg.keys():
            kwarg['baseline'] = 0
        if 'epoch' not in kwarg.keys():
            kwarg['epoch'] = 100
        if 'batch_size' not in kwarg.keys():
            kwarg['batch_size'] = 1
        if kwarg['batch_size'] > min(dataset.tr_batch_num, dataset.val_batch_num, dataset.te_batch_num):
            kwarg['batch_size'] = min(dataset.tr_batch_num, dataset.val_batch_num, dataset.te_batch_num)
            print('WARNING: batch_size should be smaller than dataset size. Automatically set batch_size =', kwarg['batch_size'])
        if 'print_type' not in kwarg.keys():
            kwarg['print_type'] = 'metric'
        if 'print_every_n_epochs' not in kwarg.keys():
            kwarg['print_every_n_epochs'] = 1
        if 'val_every_n_epochs' not in kwarg.keys():
            kwarg['val_every_n_epochs'] = 1
        if 'lr_annealing' not in kwarg.keys():
            kwarg['lr_annealing'] = 'constant'
        if 'lr_annealing_step_length' not in kwarg.keys():
            kwarg['lr_annealing_step_length'] = int(kwarg['epoch']/4)
        if 'lr_annealing_step_divisor' not in kwarg.keys():
            kwarg['lr_annealing_step_divisor'] = 2.0
        if 'early_stop' not in kwarg.keys():
            kwarg['early_stop'] = None
        if 'repeat' not in kwarg.keys():
            kwarg['repeat'] = 1
        if 'rstat' not in kwarg.keys():
            kwarg['rstat'] = 'pred'
        if 'verbose' not in kwarg.keys():
            kwarg['verbose'] = 0
        if 'validate_on' not in kwarg.keys():
            kwarg['validate_on'] = 'val'
        if 'higher_better' not in kwarg.keys():
            kwarg['higher_better'] = False
        if 'unzip' not in kwarg.keys():
            kwarg['unzip'] = 'cartesian'

        repeat_name_set = []
        repeat_metric_set = []
        repeat_time_set = []

        for _repeat in range(kwarg['repeat']):
            if not os.path.exists('model/'):
                os.mkdir('model/')
            while 1:
                self.version = time.strftime('%Y%m%d_%H%M%S')
                try:
                    model_path = 'model/' + kwarg['model_name'] + self.version
                    os.mkdir(model_path)
                    if self.summary_merged is not None:
                        self.summary_writer = tf.summary.FileWriter(model_path + '/log', self.sess.graph)
                        summary_counter = 0
                    break
                except FileExistsError:
                    time.sleep(1)
            lr = float(kwarg['lr'])
            if kwarg['higher_better']:
                performance_recorder = -1e10
            else:
                performance_recorder = 1e10
            epoch_recorder = 1e10
            start_time = time.time()

            # Reset parameters for repeat
            if kwarg['repeat'] > 1:
                self.init_parameter(seed=_repeat)
                if kwarg['verbose'] < 1:
                    print('Repeat ', _repeat, 'Start.')

            # Start training
            for epoch in range(kwarg['epoch']):
                # update an epoch
                train_ls = []
                train_me = [[] for _ in range(len(self.metric))]
                for _ in range(dataset.tr_batch_num//kwarg['batch_size']):
                    batch = dataset.tr_get_batch(kwarg['batch_size'])
                    if type(self.input) is list:
                        feed_dict = {self.input[i]: batch[i] for i in range(len(self.input))}
                        feed_dict.update({self.learning_rate: lr, self.training: True,
                             self.training_process: float((epoch+1)/kwarg['epoch'])})
                    else:
                        feed_dict = {self.learning_rate: lr, self.training: True, self.training_process: float((epoch+1)/kwarg['epoch']), self.input: batch}
                    # _re is a list of list. _re[0] is list of self.loss; _re[1+k] is list of metric[k];
                    # _re[1+K] is list of summary record (if defined). 
                    _re = self.update(feed_dict, epoch, unzip=kwarg['unzip'])
                    ls = np.mean(_re[0])
                    train_ls.append(ls)
                    for _i in range(len(self.metric)):
                        me = np.mean(_re[1+_i])
                        train_me[_i].append(me)
                    if self.summary_merged is not None:
                        for x in _re[1+len(self.metric)]:
                            self.summary_writer.add_summary(x, summary_counter)
                            summary_counter += 1
                    
                # print log
                if (epoch+1)%kwarg['print_every_n_epochs'] == 0 and kwarg['verbose'] < 1:
                    train_ls = np.mean(train_ls)
                    train_me = [np.mean(m) for m in train_me]
                    if kwarg['print_type']=='metric':
                        for _i in range(len(self.metric)):
                            print("[%s] Epoch %4d / %4d, Train %s %.5g" % (kwarg['model_name']+self.version, epoch, kwarg['epoch'], self.metric_name[_i], train_me[_i]))
                    elif kwarg['print_type']=='loss':
                        print("[%s] Epoch %4d / %4d, Train loss %.5g" % (kwarg['model_name']+self.version, epoch, kwarg['epoch'], train_ls))
                    else:
                        print("[%s] Epoch %4d / %4d, Train loss %.5g" % (kwarg['model_name']+self.version, epoch, kwarg['epoch'], train_ls))
                        for _i in range(len(self.metric)):
                            print("[%s] Epoch %4d / %4d, Train %s %.5g" % (kwarg['model_name']+self.version, epoch, kwarg['epoch'], self.metric_name[_i], train_me[_i]))
                
                # learning_rate_decay
                if kwarg['lr_annealing']=='step':
                    if (epoch+1)%kwarg['lr_annealing_step_length'] == 0:
                            lr = lr/kwarg['lr_annealing_step_divisor']
                            if kwarg['verbose'] < 1:
                                print("[%s] Epoch %4d / %4d, Learning rate decay to %.5g" % (kwarg['model_name']+self.version, epoch, kwarg['epoch'], lr))
                elif kwarg['lr_annealing']=='cosine':
                    lr = float(kwarg['lr'])*(np.cos(epoch/kwarg['epoch']*np.pi)+1.0)/2

                # validate
                if (epoch+1)%kwarg['val_every_n_epochs'] == 0:
                    # validate
                    val_me = self.test(dataset, mode=kwarg['validate_on'], target='metric', batch_size=kwarg['batch_size'], unzip=kwarg['unzip'])
                    if kwarg['verbose'] < 1:
                        for _i,_x in enumerate(val_me):
                            print("[%s] Epoch %4d / %4d, Val %s %.5g" % (kwarg['model_name']+self.version, epoch, kwarg['epoch'], self.metric_name[_i], _x))

                    # save model
                    target = val_me[0]
                    if (target < performance_recorder) ^ kwarg['higher_better']:
                        performance_recorder = target
                        epoch_recorder = epoch
                        self.saver.save(self.sess,'model/'+kwarg['model_name']+self.version+'/model')
                        if kwarg['verbose'] < 1:
                            print("[%s] Epoch %4d / %4d, Model Save Success. New record %f" % (kwarg['model_name']+self.version, epoch, kwarg['epoch'], target))

                    # Early stop
                    if kwarg['early_stop']:
                        if epoch >= epoch_recorder + kwarg['early_stop']:
                            if kwarg['verbose'] < 1:
                                print('Early stop.')
                            break

                # Customized Task
                self.iter_do(epoch)
            
            # Final Test
            self.load('model/'+kwarg['model_name']+self.version+'/model')
            test_me = self.test(dataset, mode='te', target='metric', batch_size=kwarg['batch_size'], unzip=kwarg['unzip'])

            repeat_metric_set.append(test_me)
            repeat_name_set.append(kwarg['model_name']+self.version)
            repeat_time_set.append(time.time()-start_time)
            if kwarg['verbose'] < 2:
                print('=======Training Summary=======')
                print('Time used: ', time.strftime('%H:%M:%S',time.gmtime(time.time()-start_time)))
                if performance_recorder < 1e10:
                    print('Best model at epoch ', epoch_recorder, ', with:')
                    print('Validation metric ', performance_recorder)
                    print('Test metric ', test_me)
                    print('Best Model saved as: ', 'model/'+kwarg['model_name']+self.version+'/model')

            # Here ends a repeat.
        
        # Compute repeat summary
        if kwarg['repeat'] > 1:
            test_ls_mean = np.mean(repeat_metric_set, 0)
            test_ls_std = np.std(repeat_metric_set, 0)
            if kwarg['rstat']=='pred':
                test_ls_ci = test_ls_std * 1.96
            elif kwarg['rstat']=='t':
                sem = test_ls_std / np.sqrt(kwarg['repeat'] - 1)
                test_ls_ci = scipy.stats.t.ppf(0.975, kwarg['repeat'] - 1) * sem
            total_time = np.sum(repeat_time_set)
            if kwarg['verbose'] < 2:
                print('=======Repeat Summary=======')
                print('Repeat: ', kwarg['repeat'])
                print('Model list: ', repeat_name_set)
                print('Time used: ', time.strftime('%H:%M:%S',time.gmtime(total_time)))
                print('Test loss: ', repeat_metric_set)
                for _i in range(len(self.metric)):
                    print('Average '+self.metric_name[_i], test_ls_mean[_i])
                    print('95 confidence interval: ', test_ls_ci[_i], '(',test_ls_mean[_i]-test_ls_ci[_i],'~',test_ls_mean[_i]+test_ls_ci[_i],')')

        if kwarg['verbose'] < 1:
            print('Done.')            
        
        return repeat_metric_set, repeat_name_set



'''
APDT: Analyzation
--------------------------

Provide off-shell data analyzation tools.

Check https://github.com/Zhiyuan-Wu/apdt for more information.
'''
from apdt.alz.utils import haversine, IDW
from apdt.alz.find_sites import find_sites_range
from apdt.alz.eventdetect import wavelet_detect
from apdt.alz.graph import mcl
from apdt.alz.prediction_alz import moving_predict
from apdt.alz.probability import kde_gauss_density_estimator, kde_find_identity_sigma
from apdt.alz.probability import kde_find_diagonal_sigma, kde_find_rule_of_thumb_sigma


# pylint: disable=E1136
import numpy as np

def moving_predict(x, func, par={}, width=None, step=10, **kwarg):
    '''apply a auto-regressive moving prediction on given series.
    Parameters
    -------
        x, ndarray
            the input time series, shaped [batch_size, time_length, feature_dims]
        func, callable
            the predictor, given a time series shaped [batch_size, observe_length@(0, ..., T-1), feature_dims], return a time series shaped [batch_size, predict_length@(T, ..., T+strides-1), predict_dim]
        par, dict, default {}
            the kwarg dictionary passed to func beside x.
        width, int, default to the length of input series.
            the observation window width T of func.
        step, int, default 10
            the number of moving steps.
        loss, str, default 'MAE'
            the loss function, available: 'MAE', 'RMSE'.
        long_input, bool, default False
            if True, a series of length width+strides instead of width will be given to func. strides have to be provided as well. this option is useful when some of future input can be considered as given.

    Returns
    -------
        ndarray
            the prediction.
        float
            the average prediction loss.
    '''
    # Parameter Check
    if 'loss' not in kwarg.keys():
        kwarg['loss'] = 'MAE'
    if 'long_input' not in kwarg.keys():
        kwarg['long_input'] = False
    if kwarg['long_input'] and 'strides' not in kwarg.keys():
        raise Exception('strides have to be given when long_input=True')

    if width is None:
        width = x.shape[1]

    N, L, D = x.shape
    _, strides, predict_dim = func(np.ones([N, width + kwarg['long_input'] * kwarg['strides'], D]), **par).shape

    r = np.array(x)
    r = np.concatenate([r, np.zeros([N, step * strides - (L - width) % (step * strides), D])], axis=1)
    pt = 0
    while pt+width<=L:
        slides = np.array(x[:, pt : pt + width + step * strides, :])
        if slides.shape[1] < width + step * strides:
            slides = np.concatenate([slides, np.zeros([N, width + step * strides - slides.shape[1], D])], axis=1)
        for s in range(step):
            obv = slides[:, s*strides:width+(s+kwarg['long_input'])*strides, :]
            pred = func(obv, **par)
            slides[:, width+s*strides:width+(s+1)*strides, :predict_dim] = pred
        r[:, pt+width:pt + width + step * strides, :] = slides[:, width: width + step * strides, :]
        pt = pt + step * strides

    residual = r[:N, width: L, :predict_dim] - x[:N, width: L, :predict_dim]
    if kwarg['loss'] == 'MAE':
        loss = np.mean(np.abs(residual))
    elif kwarg['loss'] == 'RMSE':
        loss = np.sqrt(np.mean(residual ** 2))
    else:
        loss = -1.0
    return r, loss


'''
APDT: Processing
--------------------------

Provide useful data processing tools for further usage.

Check https://github.com/Zhiyuan-Wu/apdt for more information.
'''
from apdt.proc.interpolate import spatial_interplot, temporal_interpolate
from apdt.proc.utils import linear_normalize, time_stamp_feature


import numpy as np

def mcl(M, expand_factor = 2, inflate_factor = 2, max_loop = 100, mult_factor = 1, eps = 1e-5):
    '''Markov Cluster Algorithm (MCL).
    MCL is a very simple clustering algorithm on graph, which intuitively believe a random walk is more likely to be trapped in a cluster. Note that in MCL a single node could be in different clusters.

    Parameters
    -----
        M, ndarray.
            The input graph adjacency matrix.
        expand_factor, int, default 2.
            The expand factor. random walk step num, larger number tends to fewer clusters.
        inflate_factor, float, default 2.
            The inflate factor. larger number tends to more clusters.
        max_loop, int, defalut 100.
            The max  number of loops to stop the algorithm.
        mult_factor, float, default 1.
            The number adds to the diagnal of matrix. tends to make algorithm stable.
        eps, float, default 1e-5.
            A small number used to check convergence.
    Returns
    -----
        clusters, dict
    Reference
    -----
        Van Dongen, S. (2000) Graph Clustering by Flow Simulation. PhD Thesis, University of Utrecht, The Netherlands.
    '''
    assert type(M) is np.ndarray and M.ndim==2
    M = M + mult_factor * np.identity(M.shape[0])
    M = M / M.sum(0)

    for i in range(max_loop):
        _M = M
        # inflate step
        M = np.power(M, inflate_factor)
        M = M / M.sum(0)
        # expend step
        M = np.linalg.matrix_power(M, expand_factor)
        # Convergence Check
        if np.abs(M - _M).max() < eps:
            break
    
    # get_cluster
    clusters = []
    for i, r in enumerate((M>0).tolist()):
        if r[i]:
            clusters.append(M[i,:]>0)

    clust_map  ={}
    for cn , c in enumerate(clusters):
        for x in  [ i for i, x in enumerate(c) if x ]:
            clust_map[cn] = clust_map.get(cn, [])  + [x]

    return M, clust_map


import numpy as np

def logloader(filepath, prefix='Average metric_0', ci=True):
    '''Load a running log file of a TFModel, and return repeat result (mean and 95CI) as a list

    Example
    -----
        r, ci = apdt.plt.logloader("slurm-190354.out")
        plt.errorbar(np.arange(len(r)), r, ci)
        plt.show()

    Parameter
    -----
        filepath: str
        prefix: str, default 'Average metric_0'
        sigma: bool, if return CI, which is considered to be at next line of prefix and startwith '95 confidence interval:  '

    Return
    -----
        list, numbers that following prefix
        list, numbers that maybe coressponding CI
    '''
    f = open(filepath, 'r')
    flag = 0
    result = []
    ci_result = []
    ci_prefix = '95 confidence interval:  '
    for line in f:
        if flag and ci and line.startswith(ci_prefix):
            ci_result.append(float(line[len(ci_prefix):].split(' ')[0]))
            flag = 0
        if line.startswith(prefix):
            result.append(float(line[len(prefix):].replace(' ','')))
            flag = 1
    
    if ci:
        return result, ci_result
    else:
        return result



'''
APDT: Data IO
--------------------------

Provide useful data fetching/loading/generating tools for standarized data structure.

Check https://github.com/Zhiyuan-Wu/apdt for more information.
'''
from apdt.io.loadnms import load_nms, load_weather
from apdt.io.fake_data import gp_data
from apdt.io.realtime import get_weather_history, get_weather_pred, real_time_weather
from apdt.io.loadfoshan import foshan_data
from apdt.io.spider import DarkSkyAPI


# pylint: disable=E1101
import numpy as np
from apdt.ml.utils import batch_norm, layer_norm

try:
    import tensorflow as tf
except:
    pass

def cnn2d_weight(**kwarg):
    '''Construct weight for 2-D CNN.
    Parameter
    ---------
        name: str
            the variable name scope to be used.
        input_dim: int, default 1
            the number of input neurals
        output_dim: int, default 128
            the number of output neurals
        input_size: list, default [28, 28]
            the input image size
        n_hidden: list of int, default 32
            the number of hidden units. if given int, it will be considered as [n_hidden] * n_layers
        conv_width: list of int, default 5
            the width of convolutional kernels. if given int, it will be considered as [conv_width] * n_layers
        n_layers: int, default 3
            the number of conv + pooling blocks
        pooling_k: int, default 1
            the pooling window size.
        trainable: bool, default True
            if these parameters are trainable.
    Return
    ------
        dict
            the dictionary of weight tensor.
    '''
    if 'name' not in kwarg:
        kwarg['name'] = 'cnn2d_weight'
    if 'input_dim' not in kwarg:
        kwarg['input_dim'] = 1
    if 'output_dim' not in kwarg:
        kwarg['output_dim'] = 128
    if 'input_size' not in kwarg:
        kwarg['input_size'] = [28, 28]
    if 'n_layers' not in kwarg:
        kwarg['n_layers'] = 3
    if 'n_hidden' not in kwarg:
        kwarg['n_hidden'] = 32
    if 'conv_width' not in kwarg:
        kwarg['conv_width'] = 5
    if 'pooling_k' not in kwarg:
        kwarg['pooling_k'] = 1
    if type(kwarg['n_hidden']) is int:
        kwarg['n_hidden'] = [kwarg['n_hidden']] * kwarg['n_layers']
    if type(kwarg['conv_width']) is int:
        kwarg['conv_width'] = [kwarg['conv_width']] * kwarg['n_layers']
    if 'trainable' not in kwarg:
        kwarg['trainable'] = True

    _init = tf.contrib.layers.xavier_initializer()
    w = {}
    with tf.variable_scope(kwarg['name']):
        w['w_0'] = tf.get_variable('w_0', [kwarg['conv_width'][0], kwarg['conv_width'][0], kwarg['input_dim'], kwarg['n_hidden'][0]], 
            initializer=_init, trainable=kwarg['trainable'])
        w['b_0'] = tf.get_variable('b_0', [kwarg['n_hidden'][0]], 
            initializer=_init, trainable=kwarg['trainable'])
        for i in range(1, kwarg['n_layers']):
            w['w_'+str(i)] = tf.get_variable('w_'+str(i), [kwarg['conv_width'][i], kwarg['conv_width'][i], kwarg['n_hidden'][i-1], kwarg['n_hidden'][i]], 
                initializer=_init, trainable=kwarg['trainable'])
            w['b_'+str(i)] = tf.get_variable('b_'+str(i), [kwarg['n_hidden'][i]], 
                initializer=_init, trainable=kwarg['trainable'])
        reduced_size = kwarg['input_size']
        for i in range(kwarg['n_layers']):
            reduced_size[0] = (reduced_size[0] + kwarg['pooling_k'] - 1) // kwarg['pooling_k']
            reduced_size[1] = (reduced_size[1] + kwarg['pooling_k'] - 1) // kwarg['pooling_k']
        reduced_size = reduced_size[0] * reduced_size[1]  * kwarg['n_hidden'][-1]
        w['out_w'] = tf.get_variable('out_w', [reduced_size, kwarg['output_dim']], 
            initializer=_init, trainable=kwarg['trainable'])
        w['out_b'] = tf.get_variable('out_b', [kwarg['output_dim']], 
            initializer=_init, trainable=kwarg['trainable'])

    return w

def cnn2d(input, weights=None, name='cnn2d', **kwarg):
    '''Map an image into one single feature vector that contains global spatial information, using 2-d CNN.
    Parameter
    ---------
        inputs: tensor
            the feature tensor to be encoded. 4d tensor. (N, W, H, D)
        weights: dict, default None
            the network weights generated by cnn1d_weight(). If given None, 
            network structure have to be clearified and corresponding weights will be newly generated.
        n_hidden: int, default 32
            the feature dimension in the model. If weights is not None, this is ignored.
        n_layers: int, default 3
            the number of conv + pooling blocks. If weights is not None, this is ignored.
        output_dim: int, default 128
            the dimension of output. If weights is not None, this is ignored.
        activation: callable, default tf.nn.relu
            the activation function on hidden nuerals.
        pooling: str, default 'MAX'
            the pooling type of hidden neural. Available: 'MAX', 'AVG'
        pooling_k: int, default 1
            the pooling window size.
        name: str
            Optional scope for `variable_scope`.
    Return
    -------
        tensor
            2d tensor. (N, D)
    '''

    if 'n_hidden' not in kwarg:
        kwarg['n_hidden'] = 32
    if 'n_layers' not in kwarg:
        kwarg['n_layers'] = 3
    if type(kwarg['n_hidden']) is int:
        kwarg['n_hidden'] = [kwarg['n_hidden']] * kwarg['n_layers']
    if 'output_dim' not in kwarg:
        kwarg['output_dim'] = 128
    if 'activation' not in kwarg:
        kwarg['activation'] = tf.nn.relu
    if 'pooling' not in kwarg:
        kwarg['pooling'] = 'MAX'
    if 'pooling_k' not in kwarg:
        kwarg['pooling_k'] = 1

    N, W, H, D = map(lambda x: x.value, input.shape)
    if weights is None:
        weights = cnn2d_weight(name=name, input_dim=D, n_hidden=kwarg['n_hidden'], n_layers=kwarg['n_layers'], output_dim=kwarg['output_dim'], input_size=[W, H], conv_width=5, pooling_k=kwarg['pooling_k'])
    else:
        kwarg['n_layers'] = len([0 for x in weights.keys() if x.startswith('w_')])

    with tf.variable_scope(name):
        h = input
        for i in range(kwarg['n_layers']):
            w = weights['w_'+str(i)]
            b = weights['b_'+str(i)]
            h = tf.nn.conv2d(input=h, filter=w, strides=[1]*4, padding='SAME') + b
            h = kwarg['activation'](h)
            h = tf.nn.pool(h, window_shape=[kwarg['pooling_k'], kwarg['pooling_k']], pooling_type=kwarg['pooling'], padding='SAME', strides=[kwarg['pooling_k'], kwarg['pooling_k']])
        h = tf.reshape(h, [N, -1])
        h = tf.matmul(h, weights['out_w']) + weights['out_b']
        
    return h

def cnn1d_weight(**kwarg):
    '''Construct weight for 1-D CNN.
    Parameter
    ---------
        out_type: str, default 'feature'
            the output type. 
            'series': output an equal-length time series.
            'feature': flatten the output series and map it to single vector using denselayer.
        name: str
            the variable name scope to be used.
        input_dim: int, default 10
            the number of input neurals
        output_dim: int, default 10
            the number of output neurals
        input_length: int, default 100
            the input seires length
        n_hidden: list of int, default 32
            the number of hidden units. if given int, it will be considered as [n_hidden] * n_layers
        conv_width: list of int, default 5
            the width of convolutional kernels. if given int, it will be considered as [conv_width] * n_layers
        n_layers: int, default 3
            the number of conv + pooling blocks
        pooling_k: int, default 2
            the pooling window size.
        trainable: bool, default True
            if these parameters are trainable.
    Return
    ------
        dict
            the dictionary of weight tensor.
    '''
    if 'out_type' not in kwarg:
        kwarg['out_type'] = 'feature'
    if 'name' not in kwarg:
        kwarg['name'] = 'cnn1d_weight'
    if 'input_dim' not in kwarg:
        kwarg['input_dim'] = 10
    if 'output_dim' not in kwarg:
        kwarg['output_dim'] = 10
    if 'input_length' not in kwarg:
        kwarg['input_length'] = 100
    if 'n_layers' not in kwarg:
        kwarg['n_layers'] = 3
    if 'n_hidden' not in kwarg:
        kwarg['n_hidden'] = 32
    if 'conv_width' not in kwarg:
        kwarg['conv_width'] = 5
    if 'pooling_k' not in kwarg:
        kwarg['pooling_k'] = 2
    if type(kwarg['n_hidden']) is int:
        kwarg['n_hidden'] = [kwarg['n_hidden']] * kwarg['n_layers']
    if type(kwarg['conv_width']) is int:
        kwarg['conv_width'] = [kwarg['conv_width']] * kwarg['n_layers']
    if 'trainable' not in kwarg:
        kwarg['trainable'] = True

    _init = tf.contrib.layers.xavier_initializer()
    w = {}
    with tf.variable_scope(kwarg['name']):
        w['w_0'] = tf.get_variable('w_0', [kwarg['conv_width'][0], kwarg['input_dim'], kwarg['n_hidden'][0]], 
            initializer=_init, trainable=kwarg['trainable'])
        w['b_0'] = tf.get_variable('b_0', [kwarg['n_hidden'][0]], 
            initializer=_init, trainable=kwarg['trainable'])
        for i in range(1, kwarg['n_layers']):
            w['w_'+str(i)] = tf.get_variable('w_'+str(i), [kwarg['conv_width'][i], kwarg['n_hidden'][i-1], kwarg['n_hidden'][i]], 
                initializer=_init, trainable=kwarg['trainable'])
            w['b_'+str(i)] = tf.get_variable('b_'+str(i), [kwarg['n_hidden'][i]], 
                initializer=_init, trainable=kwarg['trainable'])
        if kwarg['out_type'] == 'feature':
            reduced_length = kwarg['input_length']
            for i in range(kwarg['n_layers']):
                reduced_length = (reduced_length + kwarg['pooling_k'] - 1) // kwarg['pooling_k']
            w['out_w'] = tf.get_variable('out_w', [reduced_length*kwarg['n_hidden'][-1], kwarg['output_dim']], 
                initializer=_init, trainable=kwarg['trainable'])
            w['out_b'] = tf.get_variable('out_b', [kwarg['output_dim']], 
                initializer=_init, trainable=kwarg['trainable'])
        elif kwarg['out_type'] == 'series':
            w['out_w'] = tf.get_variable('out_w', [1, kwarg['n_hidden'][-1], kwarg['output_dim']], 
                initializer=_init, trainable=kwarg['trainable'])
            w['out_b'] = tf.get_variable('out_b'+str(i), [kwarg['output_dim']], 
                initializer=_init, trainable=kwarg['trainable'])

    return w

def cnn1d(input, weights=None, name='cnn1d', **kwarg):
    '''Map a time series into one single feature vector that contains global temporal information, using 1-d CNN.
    Parameter
    ---------
        inputs: tensor
            the feature tensor to be encoded. 3d tensor. (N, T, D)
        weights: dict, default None
            the network weights generated by cnn1d_weight(). If given None, 
            network structure have to be clearified and corresponding weights will be newly generated.
        out_type: str, default 'feature'
            the output type. 
            'series': output an equal-length time series. pooling should also be set to 'NONE'.
            'feature': flatten the output series and map it to single vector using denselayer.
        n_hidden: int, default 32
            the feature dimension in the model. If weights is not None, this is ignored.
        n_layers: int, default 3
            the number of conv + pooling blocks. If weights is not None, this is ignored.
        output_dim: int, default 128
            the dimension of output. If weights is not None, this is ignored.
        activation: callable, default tf.nn.relu
            the activation function on hidden nuerals.
        pooling: str, default 'MAX'
            the pooling type of hidden neural. Available: 'MAX', 'AVG', 'NONE'
            if given 'NONE', no pooling will be performed, corresponding weight should use pooling_k=1
        pooling_k: int, default 2
            the pooling window size.
        dialations: int, default None
            if set to larger than 1, there will be k-1 skipped cells between each filter element on that dimension.
            if given a list, they will be apllied layer-wise.
            only avaliable in tf==1.15.0, otherwise set this to None and disable it.
        name: str
            Optional scope for `variable_scope`.
    Return
    -------
        tensor
            2d tensor. (N, D) if out_type is 'feature'
    '''
    if 'out_type' not in kwarg:
        kwarg['out_type'] = 'feature'
    if 'n_hidden' not in kwarg:
        kwarg['n_hidden'] = 32
    if 'n_layers' not in kwarg:
        kwarg['n_layers'] = 3
    if type(kwarg['n_hidden']) is int:
        kwarg['n_hidden'] = [kwarg['n_hidden']] * kwarg['n_layers']
    if 'output_dim' not in kwarg:
        kwarg['output_dim'] = 128
    if 'activation' not in kwarg:
        kwarg['activation'] = tf.nn.relu
    if 'pooling' not in kwarg:
        kwarg['pooling'] = 'MAX'
    if 'pooling_k' not in kwarg:
        kwarg['pooling_k'] = 2
    if 'dialations' not in kwarg:
        kwarg['dialations'] = None

    N, T, D = map(lambda x: x.value, input.shape)
    if weights is None:
        weights = cnn1d_weight(name=name, input_dim=D, n_hidden=kwarg['n_hidden'], n_layers=kwarg['n_layers'], output_dim=kwarg['output_dim'], input_length=T, conv_width=5, pooling_k=kwarg['pooling_k'])
    else:
        kwarg['n_layers'] = len([0 for x in weights.keys() if x.startswith('w_')])
    if type(kwarg['dialations']) is int:
        kwarg['dialations'] = [kwarg['dialations']]*kwarg['n_layers']

    with tf.variable_scope(name):
        h = input
        for i in range(kwarg['n_layers']):
            w = weights['w_'+str(i)]
            b = weights['b_'+str(i)]
            if kwarg['dialations'] is not None:
                h = tf.nn.conv1d(h, w, 1, 'SAME',dilations=kwarg['dialations'][i]) + b
            else:
                h = tf.nn.conv1d(h, w, 1, 'SAME') + b
            h = kwarg['activation'](h)
            if kwarg['pooling'] != 'NONE':
                h = tf.nn.pool(h, window_shape=[kwarg['pooling_k']], pooling_type=kwarg['pooling'], padding='SAME', strides=[kwarg['pooling_k']])
        if kwarg['out_type'] == 'feature':
            h = tf.reshape(h, [N, -1])
            h = tf.matmul(h, weights['out_w']) + weights['out_b']
        elif kwarg['out_type'] == 'series':
            h = tf.nn.conv1d(h, weights['out_w'], 1, 'SAME') + weights['out_b']
        
    return h

def Transformer_weight(**kwarg):
    '''Construct weight for Transformer.
    Parameter
    ---------
        name: str
            the variable name scope to be used.
        input_dim: int, default 10
            the number of input neurals
        output_dim: int, default 10
            the number of output neurals
        n_hidden: list of int, default 128
            the number of hidden units
        n_layers: int, default 3
            the number of multihead-attention + dense blocks
        trainable: bool, default True
            if these parameters are trainable.
    Return
    ------
        dict
            the dictionary of weight tensor.
    '''
    if 'name' not in kwarg:
        kwarg['name'] = 'Transformer_weight'
    if 'input_dim' not in kwarg:
        kwarg['input_dim'] = 10
    if 'output_dim' not in kwarg:
        kwarg['output_dim'] = 10
    if 'n_hidden' not in kwarg:
        kwarg['n_hidden'] = 128
    if 'n_layers' not in kwarg:
        kwarg['n_layers'] = 3
    if 'trainable' not in kwarg:
        kwarg['trainable'] = True

    _init = tf.contrib.layers.xavier_initializer()
    w = {}
    with tf.variable_scope(kwarg['name']):
        w['input_project_w'] = tf.get_variable('input_project_w', [1, kwarg['input_dim'], kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
        w['input_project_b'] = tf.get_variable('input_project_b', [kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
        for i in range(kwarg['n_layers']):
            w['w_q_'+str(i)] = tf.get_variable('w_q_'+str(i), [1, kwarg['n_hidden'], kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
            w['b_q_'+str(i)] = tf.get_variable('b_q_'+str(i), [kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
            w['w_k_'+str(i)] = tf.get_variable('w_k_'+str(i), [1, kwarg['n_hidden'], kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
            w['b_k_'+str(i)] = tf.get_variable('b_k_'+str(i), [kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
            w['w_v_'+str(i)] = tf.get_variable('w_v_'+str(i), [1, kwarg['n_hidden'], kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
            w['b_v_'+str(i)] = tf.get_variable('b_v_'+str(i), [kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
            w['w_o_'+str(i)] = tf.get_variable('w_o_'+str(i), [1, kwarg['n_hidden'], kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
            w['b_o_'+str(i)] = tf.get_variable('b_o_'+str(i), [kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
            w['dense_w1_'+str(i)] = tf.get_variable('dense_w1_'+str(i), [1, kwarg['n_hidden'], kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
            w['dense_b1_'+str(i)] = tf.get_variable('dense_b1_'+str(i), [kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
            w['dense_w2_'+str(i)] = tf.get_variable('dense_w2_'+str(i), [1, kwarg['n_hidden'], kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
            w['dense_b2_'+str(i)] = tf.get_variable('dense_b2_'+str(i), [kwarg['n_hidden']], 
            initializer=_init, trainable=kwarg['trainable'])
        w['out_w'] = tf.get_variable('out_w', [1, kwarg['n_hidden'], kwarg['output_dim']], 
            initializer=_init, trainable=kwarg['trainable'])
        w['out_b'] = tf.get_variable('out_b', [kwarg['output_dim']], 
            initializer=_init, trainable=kwarg['trainable'])
    return w

def scaled_dot_product_attention(Q, K, V, causality=False,
                                 name="scaled_dot_product_attention"):
    '''Compute dot production attention. H = softmax(QK^T/sqrt(d_k))V
    Parameter
    ---------
        Q: tensor
            Packed queries. 3d tensor. [N, T_q, d_k].
        K: tensor
            Packed keys. 3d tensor. [N, T_k, d_k].
        V: tensor
            Packed values. 3d tensor. [N, T_k, d_v].
        causality: bool, default False
            If True, applies masking for future blinding. A query wont use keys and values after its index.
        name: str
            the variable name scope.
    Return
    ------
        tensor
            attention result. 3d tensor. [N, T_q, d_v]. 
    '''
    with tf.variable_scope(name):
        outputs = tf.matmul(Q, tf.transpose(K, [0, 2, 1]))  # (N, T_q, T_k)
        outputs /= (Q.shape[-1].value) ** 0.5
        if causality:
            diag_vals = tf.ones_like(outputs[0, :, :])
            tril = tf.linalg.LinearOperatorLowerTriangular(diag_vals).to_dense()
            outputs = outputs * tril - 1e10 * (1 - tril)
        outputs = tf.nn.softmax(outputs)
        outputs = tf.matmul(outputs, V)  # (N, T_q, d_v)
    return outputs

def multihead_attention(Q, K, V, w_q, w_k, w_v, w_o,
                                 b_q, b_k, b_v, b_o,
                                 causality=False,
                                 num_head=1,
                                 name="multihead_attention"):
    '''Compute multi-head attention. See Figure 2 in paper Attention is all you need.
    Parameter
    ---------
        Q: tensor
            Packed queries. 3d tensor. [N, T_q, d_k].
        K: tensor
            Packed keys. 3d tensor. [N, T_k, d_k].
        V: tensor
            Packed values. 3d tensor. [N, T_k, d_v].
        w_q, w_k, w_v, w_o: tensor
            weight for linear projection. 3d tensor. [1, d_k, d_k] (d_v for w_v). the first dimension 1 is not ignorable because we use conv1D for dense mapping.
        b_q, b_k, b_v, b_o: tensor
            bias for linear projection. 1d tensor. [d_k] (d_v for w_v).
        causality: bool, default False
            If True, applies masking for future blinding. A query wont use keys and values after its index.
        num_head: int, default 1
            The number of heads. d_k and d_v should be exact divided by num_head.
        name: str
            the variable name scope.
    Return
    ------
        tensor
            attention result. 3d tensor. [N, T_q, d_v]. 
    '''
    with tf.variable_scope(name):
        Q_ = tf.nn.conv1d(Q, w_q, 1, 'SAME') + b_q
        K_ = tf.nn.conv1d(K, w_k, 1, 'SAME') + b_k
        V_ = tf.nn.conv1d(V, w_v, 1, 'SAME') + b_v
        Q_ = tf.concat(tf.split(Q_, num_head, axis=2), axis=0) # (h*N, T_q, d_model/h)
        K_ = tf.concat(tf.split(K_, num_head, axis=2), axis=0) # (h*N, T_k, d_model/h)
        V_ = tf.concat(tf.split(V_, num_head, axis=2), axis=0) # (h*N, T_k, d_model/h)
        outputs = scaled_dot_product_attention(Q_, K_, V_, causality=causality) # (h*N, T_q, d_model/h)
        outputs = tf.concat(tf.split(outputs, num_head, axis=0), axis=2) # (N, T_q, d_model)
        outputs = tf.nn.conv1d(outputs, w_o, 1, 'SAME') + b_o
    return outputs

def positional_encoding(inputs, scope="positional_encoding"):
    '''Sinusoidal Positional_Encoding. See 3.5 in paper Attention is all you need.
    Parameter
    ---------
        inputs: tensor
            the feature tensor to be encoded. 3d tensor. (N, T, E)
        scope: str
            Optional scope for `variable_scope`.
    Return
    -------
        tensor
            3d tensor that has the same shape as inputs.
    '''
    E = inputs.get_shape().as_list()[-1] # static
    N, T = tf.shape(inputs)[0], tf.shape(inputs)[1] # dynamic
    maxlen = inputs.shape[1].value
    with tf.variable_scope(scope, reuse=tf.AUTO_REUSE):
        # position indices
        position_ind = tf.tile(tf.expand_dims(tf.range(T), 0), [N, 1]) # (N, T)

        # First part of the PE function: sin and cos argument
        position_enc = np.array([
            [pos / np.power(10000, (i-i%2)/E) for i in range(E)]
            for pos in range(maxlen)])

        # Second part, apply the cosine to even columns and sin to odds.
        position_enc[:, 0::2] = np.sin(position_enc[:, 0::2])  # dim 2i
        position_enc[:, 1::2] = np.cos(position_enc[:, 1::2])  # dim 2i+1
        position_enc = tf.convert_to_tensor(position_enc, tf.float32) # (maxlen, E)

        # lookup
        outputs = tf.nn.embedding_lookup(position_enc, position_ind)

        return tf.to_float(outputs)

def Transformer(x, weights=None, name='Transformer', **kwarg):
    '''Apply Transformer on a single series auto-regressively.
    Parameter
    ---------
        x: tensor
            the feature tensor to be encoded. 3d tensor. (N, T, D)
        weights: dict, default None
            the network weights generated by Transformer_ar_decoder_weight(). If given None, 
            network structure have to be clearified and corresponding weights will be newly generated.
        n_hidden: int, default 128
            the feature dimension in the model. If weights is not None, this is ignored.
        n_layers: int, default 3
            the number of multihead-attention + dense blocks. If weights is not None, this is ignored.
        n_head: int, default 1
            the number of heads in multihead-attention. If weights is not None, this is ignored.
        output_dim: int, default 128
            the dimension of output. If weights is not None, this is ignored.
        positional_encoding: bool, default True
            if True, the positional_encoding given by Sec.3.5 in paper Attention is all you need will be used.
        causality: bool, default False
            if True, transform at one position will not use information after its index.
        ln: int, default 1
            0: disable layer norm; 1: apply layer norm over feature dimension; 2: apply layer norm over spatial & feature dimension.
        name: str
            Optional scope for `variable_scope`.
    Return
    -------
        tensor
            3d tensor that has the same shape as inputs.
    ''' 

    if 'n_hidden' not in kwarg:
        kwarg['n_hidden'] = 128
    if 'n_layers' not in kwarg:
        kwarg['n_layers'] = 3
    if 'n_head' not in kwarg:
        kwarg['n_head'] = 4
    if 'output_dim' not in kwarg:
        kwarg['output_dim'] = 128
    if 'positional_encoding' not in kwarg:
        kwarg['positional_encoding'] = True
    if 'causality' not in kwarg:
        kwarg['causality'] = False
    if 'ln' not in kwarg:
        kwarg['ln'] = 1

    if weights is None:
        weights = Transformer_weight(name=name, input_dim=x.shape[2].value, n_hidden=kwarg['n_hidden'], n_layers=kwarg['n_layers'], output_dim=kwarg['output_dim'])
    else:
        kwarg['n_layers'] = len([0 for x in weights.keys() if x.startswith('w_q_')])

    with tf.variable_scope(name):
        x = tf.nn.conv1d(x, weights['input_project_w'], 1, 'SAME') + weights['input_project_b']
        if kwarg['positional_encoding']:
            x = x + positional_encoding(x) # its crazy to add these up, but original paper did
        for i in range(kwarg['n_layers']):
            h = multihead_attention(x, x, x,
                                    weights['w_q_'+str(i)], weights['w_k_'+str(i)], weights['w_v_'+str(i)], weights['w_o_'+str(i)],
                                    weights['b_q_'+str(i)], weights['b_k_'+str(i)], weights['b_v_'+str(i)], weights['b_o_'+str(i)],
                                    causality=kwarg['causality'], num_head=kwarg['n_head'], name="multihead_attention"+str(i))
            h = h + x
            if kwarg['ln']:
                h = layer_norm(h, K=kwarg['ln'], learnable=True, name='ln1_'+str(i))
            x = tf.nn.conv1d(h, weights['dense_w1_'+str(i)], 1, 'SAME') + weights['dense_b1_'+str(i)]
            x = tf.nn.relu(x)
            x = tf.nn.conv1d(x, weights['dense_w2_'+str(i)], 1, 'SAME') + weights['dense_b2_'+str(i)]
            x = x + h
            if kwarg['ln']:
                x = layer_norm(x, K=kwarg['ln'], learnable=True, name='ln2_'+str(i))
        o = tf.nn.conv1d(x, weights['out_w'], 1, 'SAME') + weights['out_b']
    return o

def lstm_weight(**kwarg):
    '''Construct weight for LSTM.
    Parameter
    ---------
        name: str
            the variable name scope to be used.
        input_dim: int, default 10
            the number of input neurals
        output_dim: int, default 10
            the number of output neurals
        n_hidden: list of int, default 128
            the number of hidden units
    Return
    ------
        dict
            the dictionary of weight tensor.
    '''
    if 'name' not in kwarg:
        kwarg['name'] = 'lstm'
    if 'input_dim' not in kwarg:
        kwarg['input_dim'] = 10
    if 'output_dim' not in kwarg:
        kwarg['output_dim'] = 10
    if 'n_hidden' not in kwarg:
        kwarg['n_hidden'] = 128
    if 'trainable' not in kwarg:
        kwarg['trainable'] = True
    
    _init = tf.contrib.layers.xavier_initializer()
    w = {}
    with tf.variable_scope(kwarg['name']):
        w['weight_r'] = tf.get_variable('weight_r',[2*kwarg['n_hidden'], kwarg['n_hidden']],initializer=_init, trainable=kwarg['trainable'])
        w['bias_r'] = tf.get_variable('bias_r',[kwarg['n_hidden']],initializer=_init, trainable=kwarg['trainable'])
        w['weight_u'] = tf.get_variable('weight_u',[2*kwarg['n_hidden'], kwarg['n_hidden']],initializer=_init, trainable=kwarg['trainable'])
        w['bias_u'] = tf.get_variable('bias_u',[kwarg['n_hidden']],initializer=_init, trainable=kwarg['trainable'])
        w['weight_c'] = tf.get_variable('weight_c',[2*kwarg['n_hidden'], kwarg['n_hidden']],initializer=_init, trainable=kwarg['trainable'])
        w['bias_c'] = tf.get_variable('bias_c',[kwarg['n_hidden']],initializer=_init, trainable=kwarg['trainable'])
        w['input_embedding'] = tf.get_variable('input_embedding',[1, kwarg['input_dim'], kwarg['n_hidden']],initializer=_init, trainable=kwarg['trainable'])
        w['output_transform'] = tf.get_variable('output_transform',[1, kwarg['n_hidden'], kwarg['output_dim']],initializer=_init, trainable=kwarg['trainable'])
    
    return w

def basic_lstm_cell(input, h, weight, name):
    with tf.variable_scope(name):
        x = tf.concat([input,h],axis=-1)
        r = tf.sigmoid(tf.matmul(x, weight['weight_r']) + weight['bias_r'])
        u = tf.sigmoid(tf.matmul(x, weight['weight_u']) + weight['bias_u'])
        x = tf.concat([input,h*r],axis=-1)
        c = tf.sigmoid(tf.matmul(x, weight['weight_c']) + weight['bias_c'])
        h_new = u*h + (1-u)*c
    return c,h_new

def LSTM(input, state=None, output_dim=None, n_hidden=None, weights=None, name='lstm', **kwarg):
    '''Map a time series into another using LSTM.

    parameter set (output_dim, n_hidden) will be inferenced from weights if it exists. However, in the case you have modified weights from mlp_weight (e.g. combine two weights together) things may comes wrong. Please also give (output_dim, n_hidden) in such case.

    Parameter
    ---------
        input: tensor
            a tensor shaped [batch_size, seq_length, feature_dim]
        state: tensor
            a tensor shaped [batch_size, n_hidden]
        output_dim: int
            the dimension of output vector.
        n_hidden: int
            the number of hidden neural in cell.
        weights: dict
            a dictionary of weights generated by apdt.ml.lstm_weight
    Return
    ------
        tensor
            The output vector, shaped [batch_size, seq_length, feature_dim_output]
    Note
    ----
        multi layer lstm is not supported now.
        customer rnn cell is not supported now.
        bidirectional lstm is not supported now.
    '''
    N = input.shape[0].value
    T = input.shape[1].value
    C = input.shape[2].value

    if weights is None:
        if output_dim is None:
            output_dim = 10
        if n_hidden is None:
            n_hidden = 128
        weights = lstm_weight(name=name,input_dim=C,output_dim=output_dim,n_hidden=n_hidden)
    else:
        if n_hidden is None:
            n_hidden = weights['bias_r'].shape[0].value

    if state is None:
        h = tf.zeros([N, n_hidden])
    else:
        h = state

    if 'activation' not in kwarg:
        kwarg['activation'] = tf.nn.relu
    if 'dropout' not in kwarg:
        kwarg['dropout'] = False
    if 'dropout_ratio' not in kwarg:
        kwarg['dropout_ratio'] = 0.5
    
    with tf.variable_scope(name):
        input_embedding = tf.nn.conv1d(input, weights['input_embedding'], 1, 'SAME')
        o_list = []
        for time_step in range(T):
            o, h = basic_lstm_cell(input_embedding[:,time_step], h, weights, name)
            o_list.append(o)
        pred = tf.stack(o_list, 1)
        pred = tf.nn.conv1d(pred, weights['output_transform'], 1, 'SAME')

    return pred, h

def mlp_weight(**kwarg):
    '''Construct weight for MLP.
    Parameter
    ---------
        name: str
            the variable name scope to be used.
        prefix: str, default ""
            the prefix to weight keys. useful when you want to conbine two different weight dict.
        input_dim: int, default 10
            the number of input neurals
        output_dim: int, default 10
            the number of output neurals
        n_hidden: list of int, default [128,128]
            the number of hidden units, the length of list decide the number of layers.
    Return
    ------
        dict
            the dictionary of weight tensor.
    '''
    if 'name' not in kwarg:
        kwarg['name'] = 'mlp'
    if 'prefix' not in kwarg:
        kwarg['prefix'] = ''
    if 'input_dim' not in kwarg:
        kwarg['input_dim'] = 10
    if 'output_dim' not in kwarg:
        kwarg['output_dim'] = 10
    if 'n_hidden' not in kwarg:
        kwarg['n_hidden'] = [128, 128]
    if 'trainable' not in kwarg:
        kwarg['trainable'] = True
    
    _init = tf.contrib.layers.xavier_initializer()
    w = {}
    with tf.variable_scope(kwarg['name']):
        hn = kwarg['input_dim']
        for i in range(len(kwarg['n_hidden'])):
            w[kwarg['prefix']+'weight'+str(i)] = tf.get_variable('weight'+str(i),[hn, kwarg['n_hidden'][i]],initializer=_init, trainable=kwarg['trainable'])
            w[kwarg['prefix']+'bias'+str(i)] = tf.get_variable('bias'+str(i),[kwarg['n_hidden'][i],],initializer=_init, trainable=kwarg['trainable'])
            hn = kwarg['n_hidden'][i]
        w[kwarg['prefix']+'weightout'] = tf.get_variable('weightout',[hn, kwarg['output_dim']],initializer=_init, trainable=kwarg['trainable'])
        # We use biasout to encode number of layers.
        w[kwarg['prefix']+'biasout'+'__LAYERNUM'+str(len(kwarg['n_hidden']))] = tf.get_variable('biasout',[kwarg['output_dim'],],initializer=_init, trainable=kwarg['trainable'])
    
    return w

def MLP(input, n_layers=None, output_dim=None, n_hidden=None, weights=None, name='mlp', **kwarg):
    '''Map a vector into another using MLP.

    parameter set (n_layers, output_dim, n_hidden) will be inferenced from weights if it exists. However, in the case you have modified weights from mlp_weight (e.g. combine two weights together) things may comes wrong. Please also give (n_layers, output_dim, n_hidden).

    Parameter
    ---------
        input: tensor
            a tensor shaped [batch_size, feature_dim]
        n_layers: int
            the number of hidden layers.
        output_dim: int
            the dimension of output vector.
        n_hidden: list of int
            the number of hidden neural in each layer.
        weights: dict
            a dictionary of weights generated by apdt.ml.mlp_weight
        activation: callable, default tf.nn.relu
            the activation used on hidden nueral.
        activation_kwarg: dict, default {}
            the additional key-word parameter for the given activation function.
        bn: bool, default False
            if apply batch norm on hidden layers. reuse not supported now.
        training_flag: tensor
            required if bn is True
        name: str
            the variable name scope to be used.
        prefix: str, default ""
            the prefix to weight keys. useful when you want to conbine two different weight dict.
    Return
    ------
        tensor
            The output vector, shaped [batch_size, feature_dim_output]
    Note
    ----
        Dropout is still not available now, because it shoule support automatic switch with training/testing condition.
        BN is not supported now.
    '''
    if 'activation' not in kwarg:
        kwarg['activation'] = tf.nn.relu
    if 'activation_kwarg' not in kwarg:
        kwarg['activation_kwarg'] = {}
    if 'dropout' not in kwarg:
        kwarg['dropout'] = False
    if 'dropout_ratio' not in kwarg:
        kwarg['dropout_ratio'] = 0.5
    if 'bn' not in kwarg:
        kwarg['bn'] = False
    if 'training_flag' not in kwarg:
        kwarg['training_flag'] = None
    if 'prefix' not in kwarg:
        kwarg['prefix'] = ''

    if weights is None:
        if n_layers is None:
            n_layers = 2
        if output_dim is None:
            output_dim = 10
        if n_hidden is None:
            n_hidden = [128]*n_layers
        if type(n_hidden) is int:
            n_hidden = [n_hidden]*n_layers
        weights = mlp_weight(name=name,input_dim=input.shape[-1].value,output_dim=output_dim,n_hidden=n_hidden,prefix=kwarg['prefix'])
    else:
        if n_layers is None:
            for _weight_key in weights.keys():
                if _weight_key.startswith(kwarg['prefix']+'biasout'+'__LAYERNUM'):
                    n_layers = int(_weight_key[len(kwarg['prefix']+'biasout'+'__LAYERNUM'):])
                    break
    
    with tf.variable_scope(name):
        x = input
        for i in range(n_layers):
            x = tf.tensordot(x, weights[kwarg['prefix']+'weight'+str(i)], 1) + weights[kwarg['prefix']+'bias'+str(i)]
            if kwarg['bn']:
                x = batch_norm(x, kwarg['training_flag'], name='bn'+str(i))
            x = kwarg['activation'](x, **kwarg['activation_kwarg'])
            if kwarg['dropout']:
                x = tf.nn.dropout(x, kwarg['dropout_ratio'])
        x = tf.tensordot(x, weights[kwarg['prefix']+'weightout'], 1) + weights[kwarg['prefix']+'biasout'+'__LAYERNUM'+str(n_layers)]
    return x

def wavenet_weight(**kwarg):
    '''Construct weight for wavenet.
    Parameter
    ---------
        name: str
            the variable name scope to be used.
    Return
    ------
        dict
            the dictionary of weight tensor.
    '''
    if 'name' not in kwarg:
        kwarg['name'] = 'wavenet'
    if 'res_channel' not in kwarg:
        kwarg['res_channel'] = 32
    if 'skip_channel' not in kwarg:
        kwarg['skip_channel'] = 16
    if 'input_dim' not in kwarg:
        kwarg['input_dim'] = 1
    if 'DilatedConvLayers' not in kwarg:
        kwarg['DilatedConvLayers'] = 4
    if 'n_hidden' not in kwarg:
        kwarg['n_hidden'] = 128
    if 'trainable' not in kwarg:
        kwarg['trainable'] = True
    if 'share_param_between_layers' not in kwarg:
        kwarg['share_param_between_layers'] = False
    if 'kernel_width' not in kwarg:
        kwarg['kernel_width'] = 3
    if 'output_channel' not in kwarg:
        kwarg['output_channel'] = 32

    with tf.variable_scope(kwarg['name']):
        res_channel = kwarg['res_channel']
        skip_channel = kwarg['skip_channel']
        output_channel = kwarg['output_channel']
        _init = tf.contrib.layers.xavier_initializer()
        w = {}
        if kwarg['share_param_between_layers']:
            k = 0
            w['kernel_l'+str(k)] = tf.get_variable('kernel_l'+str(k),[kwarg['kernel_width'], res_channel, 2*res_channel],initializer=_init, trainable=kwarg['trainable'])
            w['1_by_1_res_l'+str(k)] = tf.get_variable('1_by_1_res_l'+str(k),[1, res_channel, res_channel],initializer=_init, trainable=kwarg['trainable'])
            w['1_by_1_skip_l'+str(k)] = tf.get_variable('1_by_1_skip_l'+str(k),[1, res_channel, skip_channel],initializer=_init, trainable=kwarg['trainable'])
            for k in range(1, kwarg['DilatedConvLayers']):
                w['kernel_l'+str(k)] = w['kernel_l'+str(0)]
                w['1_by_1_res_l'+str(k)] = w['1_by_1_res_l'+str(0)]
                w['1_by_1_skip_l'+str(k)] = w['1_by_1_skip_l'+str(0)]
        else:
            for k in range(0, kwarg['DilatedConvLayers']):
                w['kernel_l'+str(k)] = tf.get_variable('kernel_l'+str(k),[kwarg['kernel_width'], res_channel, 2*res_channel],initializer=_init, trainable=kwarg['trainable'])
                w['1_by_1_res_l'+str(k)] = tf.get_variable('1_by_1_res_l'+str(k),[1, res_channel, res_channel],initializer=_init, trainable=kwarg['trainable'])
                w['1_by_1_skip_l'+str(k)] = tf.get_variable('1_by_1_skip_l'+str(k),[1, res_channel, skip_channel],initializer=_init, trainable=kwarg['trainable'])
        w['1_by_1_x'] = tf.get_variable('1_by_1_x', [1, kwarg['input_dim'], res_channel],initializer=_init, trainable=kwarg['trainable'])
        w['1_by_1_skip1'] = tf.get_variable(
                    '1_by_1_skip1', [1, skip_channel*kwarg['DilatedConvLayers']+kwarg['input_dim'], kwarg['n_hidden']],
                    initializer=_init, trainable=kwarg['trainable'])
        w['1_by_1_skip2'] = tf.get_variable(
                    '1_by_1_skip2', [1, kwarg['n_hidden'], output_channel],
                    initializer=_init, trainable=kwarg['trainable'])
        return w

def _WaveNet(input, weights, name, **kwarg):
    '''Map a time series into another using WaveNet.

    This implementation has been depracted and will be removed in future version. Please check WaveNet().

    Parameter
    ---------
        input: tensor
            a tensor shaped [batch_size, time_length, feature_dim]
        weights: dict
            a dictionary of weights generated by apdt.ml.wavenet_weight
        name:
            the variable name scope to be used.
    Return
    ------
        tensor
            The corresponding prediction series. a tensor shaped [batch_size, time_length, 1]
        tensor
            The loss function.
    '''
    if 'res_channel' not in kwarg.keys():
        kwarg['res_channel'] = 32
    if 'skip_channel' not in kwarg.keys():
        kwarg['skip_channel'] = 16
    if 'input_dim' not in kwarg.keys():
        kwarg['input_dim'] = 1
    if 'DilatedConvLayers' not in kwarg.keys():
        kwarg['DilatedConvLayers'] = 4
    if 'n_hidden' not in kwarg.keys():
        kwarg['n_hidden'] = 128
    if 'dilated' not in kwarg.keys():
        kwarg['dilated'] = 3
    if 'loss' not in kwarg.keys():
        kwarg['loss'] = 'MAE'
    if 'bits' not in kwarg.keys():
        kwarg['bits'] = 5
    if 'return_type' not in kwarg.keys():
        kwarg['return_type'] = 'pred+loss'

    def CasualResUnit(x,k,res_channel=kwarg['res_channel'],skip_channel=kwarg['skip_channel'],name='CasualResUnit'):
        k = int(np.log(k)/np.log(kwarg['dilated']))
        w = weights['kernel_l'+str(k)]
        _x = tf.pad(x,[[0,0],[2,0],[0,0]])
        y = tf.nn.conv1d(_x, w, 1, 'VALID')
        h,g = tf.split(y, 2, axis=-1)
        h = tf.tanh(h)
        g = tf.sigmoid(g)
        h = h*g
        w = weights['1_by_1_res_l'+str(k)]
        o = tf.nn.conv1d(h, w, 1, 'SAME')
        o = o+x
        w = weights['1_by_1_skip_l'+str(k)]
        skip = tf.nn.conv1d(h, w, 1, 'SAME')
        return o,skip
    
    def DilatedConv(x,dilated=kwarg['dilated'],res_channel=kwarg['res_channel'],skip_channel=kwarg['skip_channel'],name='DilatedConv'):
        n = x.shape[-1].value
        l = x.shape[1].value
        if l%dilated != 0:
            print('Dilated Error, when dilated at '+str(dilated))
            exit()
        num = l//dilated
        with tf.variable_scope(name) as scope:
            x = tf.reshape(x,[-1,num,dilated,n])
            _out = []
            _skip = []
            for i in range(dilated):
                out, skip = CasualResUnit(x[:,:,i],dilated,res_channel=res_channel,skip_channel=skip_channel,name='CasualResUnit#'+str(i))
                _out.append(out)
                _skip.append(skip)
                if i==0:
                    scope.reuse_variables()
            o = tf.stack(_out,axis=2)
            o = tf.reshape(o,[-1,l,n])
            skip = tf.stack(_skip,axis=2)
            skip = tf.reshape(skip,[-1,l,skip_channel])
            return o,skip
    
    def u_law_encoder(x,bits=5):
        bits = 2**bits
        x = tf.clip_by_value(x,0.0,0.99999)
        x = tf.floor(tf.log(1+bits*x)*bits/np.log(1+bits))
        x = tf.one_hot(tf.cast(tf.squeeze(x,-1),tf.int32),depth=bits)
        return x
    
    def u_law_decoder(x,bits=5):
        x = tf.cast(x,tf.float32)
        x = (tf.exp((x+0.5)/(2**bits)*np.log(1.0+(2**bits)))-1.0)/(2**bits)
        return tf.expand_dims(x,-1)
    
    def SingleChannelNetwork(x,channel,bits,encoder,decoder,name='Channel'):
        with tf.variable_scope(name):
            y = x[:,1:,channel:channel+1]
            x = x[:,:-1,:]
            _skip = [x]
            w = weights['1_by_1_x']
            x = tf.nn.conv1d(x, w, 1, 'SAME')
            for i in range(kwarg['DilatedConvLayers']):
                if i==0:
                    o,skip = DilatedConv(x,dilated=3**i,name='DilatedConv'+str(i))
                    _skip.append(skip)
                else:
                    o,skip = DilatedConv(o,dilated=3**i,name='DilatedConv'+str(i))
                    _skip.append(skip)
            skip = tf.concat(_skip,axis=-1)
            skip = tf.nn.relu(skip)
            skip_raw = skip
            w = weights['1_by_1_skip1']
            skip = tf.nn.conv1d(skip, w, 1, 'SAME')
            skip = tf.nn.relu(skip)
            w = weights['1_by_1_skip2']
            pred = tf.nn.conv1d(skip, w, 1, 'SAME')
            pred_prob = tf.nn.softmax(pred)
            pred = tf.nn.conv1d(pred_prob, decoder(tf.constant(np.arange(2**bits).reshape(1,2**bits))), 1, 'SAME')
            if kwarg['loss']=='MAE':
                loss = tf.reduce_mean(tf.abs(pred-y))
            elif kwarg['loss']=='RMSE':
                loss = tf.sqrt(tf.reduce_mean(tf.square(pred-y)))
            return pred,loss,skip_raw
    
    pred,loss,skip_raw = SingleChannelNetwork(input,0,kwarg['bits'],u_law_encoder,u_law_decoder,name=name)
    if kwarg['return_type'] == 'pred+loss':
        return pred,loss
    elif kwarg['return_type'] == 'feature':
        return skip_raw

def u_law_encoder(x,bits=5):
    bits = 2**bits
    x = tf.clip_by_value(x,0.0,0.99999)
    x = tf.floor(tf.log(1+bits*x)*bits/np.log(1+bits))
    x = tf.one_hot(tf.cast(tf.squeeze(x,-1),tf.int32),depth=bits)
    return x

def u_law_decoder(x,bits=5):
    x = tf.cast(x,tf.float32)
    x = (tf.exp((x+0.5)/(2**bits)*np.log(1.0+(2**bits)))-1.0)/(2**bits)
    return tf.expand_dims(x,-1)

def WaveNet(input, weights=None, name='WaveNet', **kwarg):
    '''Causally map a time series into another using WaveNet.
    Parameter
    ---------
        input: tensor
            a tensor shaped [batch_size, time_length, feature_dim]
        weights: dict, default None
            a dictionary of weights generated by apdt.ml.wavenet_weight. If given None, network parameters should be further clearified.
        name: str, default 'WaveNet'
            the variable name scope to be used.
        input_dim: int, default input.shape[-1]
            the dimension of input feature.
        res_channel: int, default 32
            the dimension of residual connection. More channel leads to better fit capability.
        skip_channel: int, default 16
            the dimension of skip connection. More channel leads to better fit capability.
        DilatedConvLayers: int, default 4
            the number of conv layers. More layer leads to larger receptive fields and better fit capability. It is highly suggested that input series length should far larger than receptive fields (=dilated**DilatedConvLayers)
        n_hidden: int, default 128
            the dimension of mlp hidden layers. More channel leads to better fit capability.
        dilated: int, default 3
            the dilated rate and conv kernel width. Larger number leads to more parameters and better fit capability.
        loss: str, optional: 'MAE' (default), 'RMSE'
            the loss function type.
        valid_loss: bool, default False
            if True, only predictions with full observations in receptive fields will be used to calculate loss.
        bits: int, default 5
            the u-law quatilization bit number. More bits leads to finner predcition resolution.
        return_type: str, optional 'pred+loss' (default), 'feature'
            the return type of function. 'feature' stands for raw causal features, 'pred' will further map it into a real number with mlp and u-law qutilization. 'loss' standis for the loss.
        batch_norm: bool, default False
            if set True, a batch normalization layer will be add into conv kernel.
        training: bool tensor
            required only when batch_norm is set True. Typically you can use training=self.training.

    Return
    ------
        tensor 'pred'
            The corresponding prediction series. a tensor shaped [batch_size, time_length, 1]
        tensor 'loss'
            The loss value.
        tensor 'feature'
            the raw causal features. a tensor shaped [batch_size, time_length, skip_channel*DilatedConvLayers+input_dim]
    '''
    if 'res_channel' not in kwarg.keys():
        kwarg['res_channel'] = 32
    if 'skip_channel' not in kwarg.keys():
        kwarg['skip_channel'] = 16
    if 'input_dim' not in kwarg.keys():
        try:
            kwarg['input_dim'] = input.shape[-1].value
        except:
            kwarg['input_dim'] = 1
    if 'DilatedConvLayers' not in kwarg.keys():
        kwarg['DilatedConvLayers'] = 4
    if 'n_hidden' not in kwarg.keys():
        kwarg['n_hidden'] = 128
    if 'dilated' not in kwarg.keys():
        kwarg['dilated'] = 3
    if 'kernel_width' not in kwarg.keys():
        kwarg['kernel_width'] = 3
    if 'loss' not in kwarg.keys():
        kwarg['loss'] = 'MAE'
    if 'bits' not in kwarg.keys():
        kwarg['bits'] = 5
    if 'return_type' not in kwarg.keys():
        kwarg['return_type'] = 'pred+loss'
    if 'batch_norm' not in kwarg.keys():
        kwarg['batch_norm'] = False
    if kwarg['batch_norm'] and 'training' not in kwarg.keys():
        Exception('training parameters is required when batch norm is enabled.')
    if 'valid_loss' not in kwarg.keys():
        kwarg['valid_loss'] = False
    
    if weights is None:
        weights = wavenet_weight(name=name, res_channel=kwarg['res_channel'], skip_channel=kwarg['skip_channel'],
            input_dim=kwarg['input_dim'], DilatedConvLayers=kwarg['DilatedConvLayers'], n_hidden=kwarg['n_hidden'],
            kernel_width=kwarg['kernel_width'],output_channel=2**kwarg['bits'])
    else:
        # Code to auto decision of parameters like kwarg['DilatedConvLayers']
        ...

    def SingleChannelNetwork(x,channel,bits,encoder,decoder,name='Channel'):
        with tf.variable_scope(name):
            x_input = x
            _skip = [x]
            w = weights['1_by_1_x']
            x = tf.nn.conv1d(x, w, 1, 'SAME')
            for i in range(kwarg['DilatedConvLayers']):
                _x = tf.pad(x,[[0,0],[kwarg['dilated']**i * (kwarg['kernel_width'] - 1),0],[0,0]])
                w = weights['kernel_l'+str(i)]
                h = tf.nn.conv2d(tf.expand_dims(_x,1), tf.expand_dims(w,0), [1,1,1,1], 'VALID', dilations=[1,1,kwarg['dilated']**i,1])
                if kwarg['batch_norm']:
                    h = tf.layers.batch_normalization(h, training=kwarg['training'])
                h,g = tf.split(h[:,0], 2, axis=-1)
                h = tf.tanh(h)
                g = tf.sigmoid(g)
                h = h*g
                w = weights['1_by_1_res_l'+str(i)]
                o = tf.nn.conv1d(h, w, 1, 'SAME')
                x = o+x
                w = weights['1_by_1_skip_l'+str(i)]
                skip = tf.nn.conv1d(h, w, 1, 'SAME')
                _skip.append(skip)

            skip = tf.concat(_skip,axis=-1)
            skip = tf.nn.relu(skip)
            skip_raw = skip
            w = weights['1_by_1_skip1']
            skip = tf.nn.conv1d(skip, w, 1, 'SAME')
            skip = tf.nn.relu(skip)
            w = weights['1_by_1_skip2']
            pred = tf.nn.conv1d(skip, w, 1, 'SAME')
            pred_prob = tf.nn.softmax(pred)
            pred = tf.nn.conv1d(pred_prob, decoder(tf.constant(np.arange(2**bits).reshape(1,2**bits))), 1, 'SAME')
            if not kwarg['valid_loss']:
                shift = 0
            else:
                shift = kwarg['dilated'] ** kwarg['DilatedConvLayers']
            residual = pred[:, shift:-1, channel] - x_input[:, 1+shift:, channel]
            if kwarg['loss']=='MAE':
                loss = tf.reduce_mean(tf.abs(residual))
            elif kwarg['loss']=='RMSE':
                loss = tf.sqrt(tf.reduce_mean(tf.square(residual)))
            return pred,loss,skip_raw
    
    pred,loss,skip_raw = SingleChannelNetwork(input,0,kwarg['bits'],u_law_encoder,u_law_decoder,name=name)
    if kwarg['return_type'] == 'pred+loss':
        return pred,loss
    elif kwarg['return_type'] == 'feature':
        return skip_raw
pass


# pylint: disable=E1101
import numpy as np
import apdt

class sin_regression():
    '''Sin regression meta dataset. Generate random functions of form f(x)=Asin(wx+phi)

    Parameters
    -----
        s_size: int, default 5, support sample number.
        q_size: int, default 20, query sample number.
        dataset_size: tuple of int, default (10000, 2000, 10000), number of functions in train/val/test set.
        A_range: tuple of int, default (0.1, 5.0), uniform random range of parameter A.
        phi_range: tuple of int, default (0.0, np.pi), uniform random range of parameter phi.
        w_range: tuple of int, default (8.0, 12.0), uniform random range of parameter w.
        *_range_test: parameter range of test set, default to be same as train/val. in case of you need train-test shift.
    '''
    def __init__(self, **kwarg):
        if "s_size" not in kwarg.keys():
            kwarg["s_size"] = 5
        if "q_size" not in kwarg.keys():
            kwarg["q_size"] = 20
        if "dataset_size" not in kwarg.keys():
            kwarg["dataset_size"] = (10000, 2000, 10000)
        if "A_range" not in kwarg.keys():
            kwarg["A_range"] = (0.1, 5.0)
        if "phi_range" not in kwarg.keys():
            kwarg["phi_range"] = (0.0, np.pi)
        if "w_range" not in kwarg.keys():
            kwarg["w_range"] = (8.0, 12.0)
        if "A_range_test" not in kwarg.keys():
            kwarg["A_range_test"] = kwarg["A_range"]
        if "phi_range_test" not in kwarg.keys():
            kwarg["phi_range_test"] = kwarg["phi_range"]
        if "w_range_test" not in kwarg.keys():
            kwarg["w_range_test"] = kwarg["w_range"]
        self.kwarg = kwarg
        self.tr_batch_num = kwarg["dataset_size"][0]
        self.val_batch_num = kwarg["dataset_size"][1]
        self.te_batch_num = kwarg["dataset_size"][2]
    
    def get_data(self, A, phi, w, batch_size):
        x_s = np.random.rand(batch_size, self.kwarg["s_size"], 1)
        y_s = A * np.sin(x_s * w + phi)
        x_q = np.random.rand(batch_size, self.kwarg["q_size"], 1)
        y_q = A * np.sin(x_q * w + phi)
        return [x_s, y_s, x_q, y_q]
    
    def tr_get_batch(self, batch_size=1):
        A = np.random.rand(batch_size, 1, 1) * (self.kwarg["A_range"][1] - self.kwarg["A_range"][0]) + self.kwarg["A_range"][0]
        phi = np.random.rand(batch_size, 1, 1) * (self.kwarg["phi_range"][1] - self.kwarg["phi_range"][0]) + self.kwarg["phi_range"][0]
        w = np.random.rand(batch_size, 1, 1) * (self.kwarg["w_range"][1] - self.kwarg["w_range"][0]) + self.kwarg["w_range"][0]
        return self.get_data(A, phi, w, batch_size)

    def val_get_batch(self, batch_size=1):
        A = np.random.rand(batch_size, 1, 1) * (self.kwarg["A_range"][1] - self.kwarg["A_range"][0]) + self.kwarg["A_range"][0]
        phi = np.random.rand(batch_size, 1, 1) * (self.kwarg["phi_range"][1] - self.kwarg["phi_range"][0]) + self.kwarg["phi_range"][0]
        w = np.random.rand(batch_size, 1, 1) * (self.kwarg["w_range"][1] - self.kwarg["w_range"][0]) + self.kwarg["w_range"][0]
        return self.get_data(A, phi, w, batch_size)

    def te_get_batch(self, batch_size=1):
        A = np.random.rand(batch_size, 1, 1) * (self.kwarg["A_range_test"][1] - self.kwarg["A_range_test"][0]) + self.kwarg["A_range_test"][0]
        phi = np.random.rand(batch_size, 1, 1) * (self.kwarg["phi_range_test"][1] - self.kwarg["phi_range_test"][0]) + self.kwarg["phi_range_test"][0]
        w = np.random.rand(batch_size, 1, 1) * (self.kwarg["w_range_test"][1] - self.kwarg["w_range_test"][0]) + self.kwarg["w_range_test"][0]
        return self.get_data(A, phi, w, batch_size)

class ContinualDataset():
    ''' A Meta Dataset for Continual Learning purpose. It first construct a large dataset, and then return subset as sub-dataset.

    Parameter
    ------
        datapack: an apdt.DataPack
        task_num: the total number of child dataset

    Method
    ------
        get_dataset: return a sub-dataset.
    '''
    def __init__(self, datapack, **kwarg):
        if 'seed' not in kwarg.keys():
            kwarg['seed'] = None
        if 'task_num' not in kwarg.keys():
            kwarg['task_num'] = 10

        self._kwarg = kwarg

        if type(datapack) is apdt.DataPack:
            self.pre_process(datapack, **kwarg)
        else:
            raise NotImplementedError

        if kwarg['task_num'] > self.tr_batch_num:
            raise Exception("No enough data to construct ", kwarg['task_num'], " sub-datasets.")

    def pre_process(self, datapack, **kwarg):
        if 'method' not in kwarg.keys():
            raise Exception("You have to re-define pre_process function or use a pre-defined one using 'method' argument.")

        if kwarg['method'] == 'window':
            if 'time-aligned' not in datapack.tag:
                raise Exception("datapack should be aligned along time axis first.")
            if 'fixed-location' not in datapack.tag:
                raise Exception("window method only support fixed station data.")
            
            if 'seq_len' not in kwarg.keys():
                kwarg['seq_len'] = 100
            if 'strides' not in kwarg.keys():
                kwarg['strides'] = kwarg['seq_len']
            if 'normalize' not in kwarg.keys():
                kwarg['normalize'] = True
            if 'normalize_method' not in kwarg.keys():
                kwarg['normalize_method'] = '99pt'
            if 'feature' not in kwarg.keys():
                kwarg['feature'] = [x for x in datapack.data.columns if x.startswith('data')]

            self.method = 'window'
            T = len(datapack.data.index.drop_duplicates()) 
            N = len(datapack.data.site_id.drop_duplicates())
            D = len(kwarg['feature'])
            if kwarg['normalize']:
                datapack = apdt.proc.linear_normalize(datapack, kwarg['feature'], kwarg['normalize_method'])
            self.data = datapack.data.reset_index().sort_values(['datetime', 'site_id'])[kwarg['feature']].values.reshape((T, N, D))
            
            tr_list = []
            self.tr_batch_num = (T - kwarg['seq_len']) // kwarg['strides'] + 1
            for i in range(self.tr_batch_num):
                tr_list.append(self.data[i*kwarg['strides']:i*kwarg['strides']+kwarg['seq_len']])
            self.tr = np.stack(tr_list)
            self.tr = np.transpose(self.tr, (0,2,1,3))

            self.seq_len = kwarg['seq_len']
            self.strides = kwarg['strides']

    def get_dataset(self, idx1, idx2=None, split_ratio=None):
        ''' Return a dataset that include samples from task idx1 (or a combined dataset from task idx1 - idx2 (include))

        Parameters
        -----
            idx1: target task index, 0<=idx1<=self.task_num
            idx2: if given, task idx1 - idx2 (include both ends) will be combined into one and returned
            split_ratio: the ratio to bu used as training set. note that this is a random split instead of split-along-time.
        '''
        sample_num_per_task = self.tr.shape[0] // self._kwarg['task_num']
        idx_start = idx1 * sample_num_per_task
        if idx2 is None:
            idx2 = idx1
        idx_end = (idx2 + 1) * sample_num_per_task
        return _ContinualDataset_child(self, idx_start, idx_end, split_ratio, self._kwarg['seed'])

class _ContinualDataset_child():
    '''A dataset object generated from ContinualDataset, provide basic interface same as apdt.DataSet, e.g., shuffle, get_batch.

    Note: This dataset use a random split of train/val instead of split-along-time
    '''
    def __init__(self, father, idx_start, idx_end, split_ratio=None, seed=None):
        '''
        Parameters
        -----
            father: ContinualDataset instance.
            idx_start: start sample index (include)
            idx_end: start sample index (exclude)
            split_ratio: the ratio of training set, rest will be used as validation. default no validation.
            seed: seed for randomness, default none.
        '''
        if split_ratio is not None:
            if seed is not None:
                np.random.seed(seed)
            perm = np.arange(idx_start, idx_end)
            np.random.shuffle(perm)
            tr_idx = perm[:int((idx_end-idx_start)*split_ratio)]
            val_idx = perm[int((idx_end-idx_start)*split_ratio):]
            self.tr = father.tr[tr_idx]
            self.val = father.tr[val_idx]
            self.te = father.tr[idx_start: idx_end]
        else:
            self.tr = father.tr[idx_start: idx_end]
            self.val = father.tr[idx_start: idx_end]
            self.te = father.tr[idx_start: idx_end]

        self.tr_batch_num = self.tr.shape[0]
        self.val_batch_num = self.val.shape[0]
        self.te_batch_num = self.te.shape[0]

        self._init_counter(seed=seed)

    def _init_counter(self, **kwarg):
        if kwarg['seed'] is not None:
            np.random.seed(kwarg['seed'])
        self.tr_batch_counter = 0
        self.val_batch_counter = 0
        self.te_batch_counter = 0
        self.tr_batch_perm = np.linspace(0, self.tr_batch_num-1, self.tr_batch_num, dtype=np.int32)
        self.val_batch_perm = np.linspace(0, self.val_batch_num-1, self.val_batch_num, dtype=np.int32)
        self.te_batch_perm = np.linspace(0, self.te_batch_num-1, self.te_batch_num, dtype=np.int32)
        np.random.shuffle(self.tr_batch_perm)

    def tr_get_batch(self, batch_size=1):
        if batch_size > self.tr_batch_num:
            batch_size = self.tr_batch_num
        if self.tr_batch_counter + batch_size > self.tr_batch_num:
            np.random.shuffle(self.tr_batch_perm)
            self.tr_batch_counter = 0
        target_index = self.tr_batch_perm[self.tr_batch_counter: self.tr_batch_counter + batch_size]
        self.tr_batch_counter = self.tr_batch_counter + batch_size
        batch = self.tr[target_index]
        batch = self.post_process(batch, mode='train')
        return batch

    def val_get_batch(self, batch_size=1):
        if batch_size > self.val_batch_num:
            batch_size = self.val_batch_num
        if self.val_batch_counter + batch_size > self.val_batch_num:
            self.val_batch_counter = 0
        target_index = self.val_batch_perm[self.val_batch_counter: self.val_batch_counter + batch_size]
        self.val_batch_counter = self.val_batch_counter + batch_size
        batch = self.val[target_index]
        batch = self.post_process(batch, mode='validate')
        return batch
    
    def te_get_batch(self, batch_size=1):
        if batch_size > self.te_batch_num:
            batch_size = self.te_batch_num
        if self.te_batch_counter + batch_size > self.te_batch_num:
            self.te_batch_counter = 0
        target_index = self.te_batch_perm[self.te_batch_counter: self.te_batch_counter + batch_size]
        self.te_batch_counter = self.te_batch_counter + batch_size
        batch = self.te[target_index]
        batch = self.post_process(batch, mode='test')
        return batch

    def post_process(self, batch, **kwarg):
        '''Define the action on post_process of batches from dataset, default doing nothing.

        Parameters
        ------
            batch, tensor
            mode, in ['train, 'validate', 'test]
        '''
        return batch